#!/usr/bin/env python3
from __future__ import annotations

from itertools import chain
from typing import Dict, Iterable, List, Optional, Union, cast

import numpy as np
import numpy.typing as npt
import pandas as pd
from scipy import sparse
from sklearn.preprocessing import KBinsDiscretizer as SK_KBinsDiscretizer

from snowflake import snowpark
from snowflake.ml._internal import telemetry
from snowflake.ml._internal.exceptions import error_codes, exceptions
from snowflake.ml.modeling.framework import base
from snowflake.snowpark import functions as F, types as T
from snowflake.snowpark._internal.utils import (
    TempObjectType,
    random_name_for_temp_object,
)

# constants used to validate the compatibility of the kwargs passed to the sklearn
# transformer with the sklearn version
_SKLEARN_INITIAL_KEYWORDS = [
    "n_bins",
    "encode",
    "strategy",
]  # initial keywords in sklearn

# TODO(tbao): decide if we want to support subsample
_SKLEARN_UNUSED_KEYWORDS = [
    "dtype",
    "subsample",
    "random_state",
]  # sklearn keywords that are unused in snowml

_SNOWML_ONLY_KEYWORDS = [
    "input_cols",
    "output_cols",
]  # snowml only keywords not present in sklearn

_VALID_ENCODING_SCHEME = ["onehot", "onehot-dense", "ordinal"]
_VALID_STRATEGY = ["uniform", "quantile", "kmeans"]


# NB: numpy doesn't work with decimal type, so converting to float
def decimal_to_float(data: npt.NDArray[np.generic]) -> npt.NDArray[np.float32]:
    return np.array([float(x) for x in data])


# TODO(tbao): support kmeans with snowpark if needed
class KBinsDiscretizer(base.BaseTransformer):
    """
    Bin continuous data into intervals.

    Args:
        n_bins: int or array-like of shape (n_features,), default=5
            The number of bins to produce. Raises ValueError if n_bins < 2.

        encode: {'onehot', 'onehot-dense', 'ordinal'}, default='onehot'
            Method used to encode the transformed result.

            - 'onehot': Encode the transformed result with one-hot encoding and return a sparse representation.
            - 'onehot-dense': Encode the transformed result with one-hot encoding and return separate column for
                each encoded value.
            - 'ordinal': Return the bin identifier encoded as an integer value.

        strategy: {'uniform', 'quantile'}, default='quantile'
            Strategy used to define the widths of the bins.

            - 'uniform': All bins in each feature have identical widths.
            - 'quantile': All bins in each feature have the same number of points.

        input_cols: str or Iterable [column_name], default=None
            Single or multiple input columns.

        output_cols: str or Iterable [column_name], default=None
            Single or multiple output columns.

        drop_input_cols: boolean, default=False
            Remove input columns from output if set True.

    Attributes:
        bin_edges_: ndarray of ndarray of shape (n_features,)
            The edges of each bin. Contain arrays of varying shapes (n_bins_, )

        n_bins_: ndarray of shape (n_features,), dtype=np.int_
            Number of bins per feature.
    """

    def __init__(
        self,
        *,
        n_bins: Union[int, List[int]] = 5,
        encode: str = "onehot",
        strategy: str = "quantile",
        input_cols: Optional[Union[str, Iterable[str]]] = None,
        output_cols: Optional[Union[str, Iterable[str]]] = None,
        drop_input_cols: Optional[bool] = False,
    ) -> None:
        super().__init__(drop_input_cols=drop_input_cols)
        self.n_bins = n_bins
        self.encode = encode
        self.strategy = strategy
        self.set_input_cols(input_cols)
        self.set_output_cols(output_cols)

    def _enforce_params(self) -> None:
        self.n_bins = self.n_bins if isinstance(self.n_bins, Iterable) else [self.n_bins] * len(self.input_cols)
        if len(self.n_bins) != len(self.input_cols):
            raise exceptions.SnowflakeMLException(
                error_code=error_codes.INVALID_ATTRIBUTE,
                original_exception=ValueError(
                    f"n_bins must have same size as input_cols, got: {self.n_bins} vs {self.input_cols}"
                ),
            )
        for idx, b in enumerate(self.n_bins):
            if b < 2:
                raise exceptions.SnowflakeMLException(
                    error_code=error_codes.INVALID_ATTRIBUTE,
                    original_exception=ValueError(f"n_bins cannot be less than 2, got: {b} at index {idx}"),
                )
        if self.encode not in _VALID_ENCODING_SCHEME:
            raise exceptions.SnowflakeMLException(
                error_code=error_codes.INVALID_ATTRIBUTE,
                original_exception=ValueError(f"encode must be one of f{_VALID_ENCODING_SCHEME}, got: {self.encode}"),
            )
        if self.strategy not in _VALID_STRATEGY:
            raise exceptions.SnowflakeMLException(
                error_code=error_codes.INVALID_ATTRIBUTE,
                original_exception=ValueError(f"strategy must be one of f{_VALID_STRATEGY}, got: {self.strategy}"),
            )

    def _reset(self) -> None:
        super()._reset()
        self.bin_edges_: Optional[npt.NDArray[np.float32]] = None
        self.n_bins_: Optional[npt.NDArray[np.int32]] = None

    @telemetry.send_api_usage_telemetry(
        project=base.PROJECT,
        subproject=base.SUBPROJECT,
    )
    def fit(self, dataset: Union[snowpark.DataFrame, pd.DataFrame]) -> KBinsDiscretizer:
        """
        Fit KBinsDiscretizer with dataset.

        Args:
            dataset: Input dataset.

        Returns:
            Fitted self instance.
        """
        self._reset()
        self._enforce_params()
        super()._check_input_cols()
        super()._check_dataset_type(dataset)

        if isinstance(dataset, pd.DataFrame):
            self._fit_sklearn(dataset)
        else:
            self._fit_snowpark(dataset)

        self._is_fitted = True
        return self

    @telemetry.send_api_usage_telemetry(
        project=base.PROJECT,
        subproject=base.SUBPROJECT,
    )
    @telemetry.add_stmt_params_to_df(
        project=base.PROJECT,
        subproject=base.SUBPROJECT,
    )
    def transform(
        self, dataset: Union[snowpark.DataFrame, pd.DataFrame]
    ) -> Union[snowpark.DataFrame, pd.DataFrame, sparse.csr_matrix]:
        """
        Discretize the data.

        Args:
            dataset: Input dataset.

        Returns:
            Discretized output data based on input type.
            - If input is snowpark DataFrame, returns snowpark DataFrame
            - If input is a pd.DataFrame and 'self.encdoe=onehot', returns 'csr_matrix'
            - If input is a pd.DataFrame and 'self.encode in ['ordinal', 'onehot-dense']', returns 'pd.DataFrame'
        """
        self._enforce_fit()
        super()._check_input_cols()
        super()._check_output_cols()
        super()._check_dataset_type(dataset)

        if isinstance(dataset, snowpark.DataFrame):
            output_df = self._transform_snowpark(dataset)
        else:
            output_df = self._transform_sklearn(dataset)

        return self._drop_input_columns(output_df) if self._drop_input_cols is True else output_df

    def _fit_snowpark(self, dataset: snowpark.DataFrame) -> None:
        if self.strategy == "quantile":
            self._handle_quantile(dataset)
        elif self.strategy == "uniform":
            self._handle_uniform(dataset)
        elif self.strategy == "kmeans":
            raise exceptions.SnowflakeMLException(
                error_code=error_codes.NOT_IMPLEMENTED,
                original_exception=NotImplementedError("kmeans not supported yet"),
            )

    def _handle_quantile(self, dataset: snowpark.DataFrame) -> None:
        """
        Compute bins with percentile values of the feature.
        All bins in each feature will have the same number of points.

        Args:
            dataset: Input dataset.
        """
        # 1. Collect percentiles for each feature column
        # NB: if SQL compilation ever take too long on wide schema, consider applying optimization mentioned in
        # https://docs.google.com/document/d/1cilfCCtKYv6HvHqaqdZxfHAvQ0gg-t1AM8KYCQtJiLE/edit
        agg_queries = []
        for idx, col_name in enumerate(self.input_cols):
            percentiles = np.linspace(0, 1, cast(List[int], self.n_bins)[idx] + 1)
            for i, pct in enumerate(percentiles.tolist()):
                agg_queries.append(F.percentile_cont(pct).within_group(col_name).alias(f"{col_name}_pct_{i}"))
        state_df = dataset.agg(agg_queries)
        state = (
            state_df.to_pandas(
                statement_params=telemetry.get_statement_params(base.PROJECT, base.SUBPROJECT, self.__class__.__name__)
            )
            .to_numpy()
            .ravel()
        )

        # 2. Populate internal state variables
        # TODO(tbao): Remove bins whose width are too small (i.e., <= 1e-8)
        self.bin_edges_ = np.zeros(len(self.input_cols), dtype=object)
        self.n_bins_ = np.zeros(len(self.input_cols), dtype=np.int_)
        start = 0
        for i, b in enumerate(cast(List[int], self.n_bins)):
            self.bin_edges_[i] = decimal_to_float(state[start : start + b + 1])
            start += b + 1
            self.n_bins_[i] = len(self.bin_edges_[i]) - 1

    def _handle_uniform(self, dataset: snowpark.DataFrame) -> None:
        """
        Compute bins with min and max value of the feature.
        All bins in each feature will have identical widths.

        Args:
            dataset: Input dataset.
        """
        # 1. Collect min and max for each feature column
        agg_queries = list(
            chain.from_iterable((F.min(x).alias(f"{x}_min"), F.max(x).alias(f"{x}_max")) for x in self.input_cols)
        )
        state_df = dataset.select(*agg_queries)
        state = (
            state_df.to_pandas(
                statement_params=telemetry.get_statement_params(base.PROJECT, base.SUBPROJECT, self.__class__.__name__)
            )
            .to_numpy()
            .ravel()
        )

        # 2. Populate internal state variables
        self.bin_edges_ = np.zeros(len(self.input_cols), dtype=object)
        self.n_bins_ = np.zeros(len(self.input_cols), dtype=np.int_)
        for i, b in enumerate(cast(List[int], self.n_bins)):
            self.bin_edges_[i] = np.linspace(state[i * 2], state[i * 2 + 1], b + 1)
            self.n_bins_[i] = len(self.bin_edges_[i]) - 1

    def _fit_sklearn(self, dataset: pd.DataFrame) -> None:
        dataset = self._use_input_cols_only(dataset)
        sklearn_discretizer = self._create_unfitted_sklearn_object()
        sklearn_discretizer.fit(dataset[self.input_cols])

        self.bin_edges_ = sklearn_discretizer.bin_edges_
        self.n_bins_ = sklearn_discretizer.n_bins_
        if "onehot" in self.encode:
            self._encoder = sklearn_discretizer._encoder

    def _create_unfitted_sklearn_object(self) -> SK_KBinsDiscretizer:
        sklearn_args = self.get_sklearn_args(
            default_sklearn_obj=SK_KBinsDiscretizer(),
            sklearn_initial_keywords=_SKLEARN_INITIAL_KEYWORDS,
            sklearn_unused_keywords=_SKLEARN_UNUSED_KEYWORDS,
            snowml_only_keywords=_SNOWML_ONLY_KEYWORDS,
        )
        return SK_KBinsDiscretizer(**sklearn_args)

    def _create_sklearn_object(self) -> SK_KBinsDiscretizer:
        sklearn_discretizer = self._create_unfitted_sklearn_object()
        if self._is_fitted:
            sklearn_discretizer.bin_edges_ = self.bin_edges_
            sklearn_discretizer.n_bins_ = self.n_bins_
            if "onehot" in self.encode:
                sklearn_discretizer._encoder = self._encoder
        return sklearn_discretizer

    def _transform_snowpark(self, dataset: snowpark.DataFrame) -> snowpark.DataFrame:
        if self.encode == "ordinal":
            return self._handle_ordinal(dataset)
        elif self.encode == "onehot":
            return self._handle_onehot(dataset)
        elif self.encode == "onehot-dense":
            return self._handle_onehot_dense(dataset)
        else:
            raise exceptions.SnowflakeMLException(
                error_code=error_codes.INVALID_ATTRIBUTE,
                original_exception=ValueError(f"{self.encode} is not a valid encoding scheme."),
            )

    def _handle_ordinal(self, dataset: snowpark.DataFrame) -> snowpark.DataFrame:
        """
        Transform dataset with bucketization and output as ordinal encoding.

        Args:
            dataset: Input dataset.

        Returns:
            Output dataset with ordinal encoding.
        """
        passthrough_columns = [c for c in dataset.columns if c not in self.output_cols]
        # NB: the reason we need to generate a random UDF name each time is because the UDF registration
        # is centralized per database, so if there are multiple sessions with same UDF name, there might be
        # a conflict and some parties could fail to fetch the UDF.
        udf_name = random_name_for_temp_object(TempObjectType.FUNCTION)

        # 1. Register vec_bucketize UDF
        @F.pandas_udf(  # type: ignore[arg-type, misc]
            is_permanent=False,
            name=udf_name,
            replace=True,
            packages=["numpy"],
            session=dataset._session,
            statement_params=telemetry.get_statement_params(base.PROJECT, base.SUBPROJECT, self.__class__.__name__),
        )
        def vec_bucketize_temp(x: T.PandasSeries[float], boarders: T.PandasSeries[List[float]]) -> T.PandasSeries[int]:
            # NB: vectorized udf doesn't work well with const array arg, so we pass it in as a list via PandasSeries
            boarders = boarders[0]
            res = np.searchsorted(boarders[1:-1], x, side="right")
            return pd.Series(res)  # type: ignore[no-any-return]

        # 2. compute bucket per feature column
        for idx, input_col in enumerate(self.input_cols):
            output_col = self.output_cols[idx]
            assert self.bin_edges_ is not None
            boarders = [F.lit(float(x)) for x in self.bin_edges_[idx]]
            dataset = dataset.select(
                *dataset.columns,
                F.call_udf(udf_name, F.col(input_col), F.array_construct(*boarders)).alias(output_col),
            )
        # Reorder columns. Passthrough columns are added at the right to the output of the transformers.
        dataset = dataset[self.output_cols + passthrough_columns]
        return dataset

    def _handle_onehot(self, dataset: snowpark.DataFrame) -> snowpark.DataFrame:
        """
        Transform dataset with bucketization and output as sparse representation:
        {"{bucket_id}": 1.0, "array_length": {num_buckets}}

        Args:
            dataset: Input dataset.

        Returns:
            Output dataset in sparse representation.
        """
        passthrough_columns = [c for c in dataset.columns if c not in self.output_cols]
        udf_name = random_name_for_temp_object(TempObjectType.FUNCTION)

        @F.pandas_udf(  # type: ignore[arg-type, misc]
            is_permanent=False,
            name=udf_name,
            replace=True,
            packages=["numpy"],
            session=dataset._session,
            statement_params=telemetry.get_statement_params(base.PROJECT, base.SUBPROJECT, self.__class__.__name__),
        )
        def vec_bucketize_sparse_output_temp(
            x: T.PandasSeries[float], boarders: T.PandasSeries[List[float]]
        ) -> T.PandasSeries[Dict[str, int]]:
            res: List[Dict[str, int]] = []
            boarders = boarders[0]
            buckets = np.searchsorted(boarders[1:-1], x, side="right")
            assert isinstance(buckets, np.ndarray), f"expecting buckets to be numpy ndarray, got {type(buckets)}"
            array_length = len(boarders) - 1
            for bucket in buckets:
                res.append({str(bucket): 1, "array_length": array_length})
            return pd.Series(res)

        for idx, input_col in enumerate(self.input_cols):
            assert self.bin_edges_ is not None
            output_col = self.output_cols[idx]
            boarders = [F.lit(float(x)) for x in self.bin_edges_[idx]]
            dataset = dataset.select(
                *dataset.columns,
                F.call_udf(udf_name, F.col(input_col), F.array_construct(*boarders)).alias(output_col),
            )
        # Reorder columns. Passthrough columns are added at the right to the output of the transformers.
        dataset = dataset[self.output_cols + passthrough_columns]
        return dataset

    def _handle_onehot_dense(self, dataset: snowpark.DataFrame) -> snowpark.DataFrame:
        """
        Transform dataset with bucketization and output as onehot dense representation:
        Each category will be reprensented in its own output column.

        Args:
            dataset: Input dataset.

        Returns:
            Output dataset in dense representation.
        """
        original_dataset_columns = dataset.columns[:]
        all_output_cols = []

        udf_name = random_name_for_temp_object(TempObjectType.FUNCTION)

        @F.pandas_udf(  # type: ignore[arg-type, misc]
            name=udf_name,
            replace=True,
            packages=["numpy"],
            session=dataset._session,
            statement_params=telemetry.get_statement_params(base.PROJECT, base.SUBPROJECT, self.__class__.__name__),
        )
        def vec_bucketize_dense_output_temp(
            x: T.PandasSeries[float], boarders: T.PandasSeries[List[float]]
        ) -> T.PandasSeries[List[int]]:
            res: List[npt.NDArray[np.int32]] = []
            boarders = boarders[0]
            buckets = np.searchsorted(boarders[1:-1], x, side="right")
            assert isinstance(buckets, np.ndarray), f"expecting buckets to be numpy ndarray, got {type(buckets)}"
            for bucket in buckets:
                encoded = np.zeros(len(boarders), dtype=int)
                encoded[bucket] = 1
                res.append(encoded)
            return pd.Series(res)

        for idx, input_col in enumerate(self.input_cols):
            assert self.bin_edges_ is not None
            output_col = self.output_cols[idx]
            boarders = [F.lit(float(x)) for x in self.bin_edges_[idx]]
            dataset = dataset.select(
                *dataset.columns,
                F.call_udf(udf_name, F.col(input_col), F.array_construct(*boarders)).alias(output_col),
            )
            dataset = dataset.with_columns(
                [f"{output_col}_{i}" for i in range(len(boarders) - 1)],
                [F.col(output_col)[i].cast(T.IntegerType()) for i in range(len(boarders) - 1)],
            ).drop(output_col)
            all_output_cols += [f"{output_col}_{i}" for i in range(len(boarders) - 1)]

        # Reorder columns. Passthrough columns are added at the right to the output of the transformers.
        dataset = dataset[all_output_cols + original_dataset_columns]
        return dataset

    def _transform_sklearn(self, dataset: pd.DataFrame) -> Union[pd.DataFrame, sparse.csr_matrix]:
        """
        Transform pandas dataframe using sklearn KBinsDiscretizer.
        Output be a sparse matrix if self.encode='onehot' and pandas dataframe otherwise.

        Args:
            dataset: Input dataset.

        Returns:
            Output dataset.
        """
        self._enforce_fit()
        encoder_sklearn = self.to_sklearn()

        transformed_dataset = encoder_sklearn.transform(dataset[self.input_cols])

        if self.encode == "ordinal":
            dataset = dataset.copy()
            dataset[self.output_cols] = transformed_dataset
            return dataset
        elif self.encode == "onehot-dense":
            dataset = dataset.copy()
            dataset[self.get_output_cols()] = transformed_dataset
            return dataset
        else:
            return transformed_dataset

    def get_output_cols(self) -> List[str]:
        """
        Get output column names.
        Expand output column names for 'onehot-dense' encoding.

        Returns:
            Output column names.
        """

        if self.encode == "onehot-dense":
            output_cols = []
            for idx, col in enumerate(self.output_cols):
                output_cols.extend([f"{col}_{i}" for i in range(self.n_bins_[idx])])  # type: ignore[index]
            return output_cols
        else:
            return self.output_cols



import os
from typing import TYPE_CHECKING, Callable, Dict, Optional, Type, cast, final

import numpy as np
import pandas as pd
from typing_extensions import TypeGuard, Unpack

from snowflake.ml._internal import type_utils
from snowflake.ml.model import custom_model, model_signature, type_hints as model_types
from snowflake.ml.model._packager.model_env import model_env
from snowflake.ml.model._packager.model_handlers import _base, _utils as handlers_utils
from snowflake.ml.model._packager.model_handlers_migrator import base_migrator
from snowflake.ml.model._packager.model_meta import (
    model_blob_meta,
    model_meta as model_meta_api,
)
from snowflake.ml.model._signatures import (
    numpy_handler,
    tensorflow_handler,
    utils as model_signature_utils,
)

if TYPE_CHECKING:
    import tensorflow


@final
class TensorFlowHandler(_base.BaseModelHandler["tensorflow.Module"]):
    """Handler for TensorFlow based model.

    Currently tensorflow.Module based classes are supported.
    """

    HANDLER_TYPE = "tensorflow"
    HANDLER_VERSION = "2023-12-01"
    _MIN_SNOWPARK_ML_VERSION = "1.0.12"
    _HANDLER_MIGRATOR_PLANS: Dict[str, Type[base_migrator.BaseModelHandlerMigrator]] = {}

    MODELE_BLOB_FILE_OR_DIR = "model"
    DEFAULT_TARGET_METHODS = ["__call__"]

    @classmethod
    def can_handle(
        cls,
        model: model_types.SupportedModelType,
    ) -> TypeGuard["tensorflow.nn.Module"]:
        return type_utils.LazyType("tensorflow.Module").isinstance(model)

    @classmethod
    def cast_model(
        cls,
        model: model_types.SupportedModelType,
    ) -> "tensorflow.Module":
        import tensorflow

        assert isinstance(model, tensorflow.Module)

        return cast(tensorflow.Module, model)

    @classmethod
    def save_model(
        cls,
        name: str,
        model: "tensorflow.Module",
        model_meta: model_meta_api.ModelMetadata,
        model_blobs_dir_path: str,
        sample_input: Optional[model_types.SupportedDataType] = None,
        is_sub_model: Optional[bool] = False,
        **kwargs: Unpack[model_types.TensorflowSaveOptions],
    ) -> None:
        import tensorflow

        assert isinstance(model, tensorflow.Module)

        if isinstance(model, tensorflow.keras.Model):
            default_target_methods = ["predict"]
        else:
            default_target_methods = cls.DEFAULT_TARGET_METHODS

        if not is_sub_model:
            target_methods = handlers_utils.get_target_methods(
                model=model,
                target_methods=kwargs.pop("target_methods", None),
                default_target_methods=default_target_methods,
            )

            def get_prediction(
                target_method_name: str, sample_input: "model_types.SupportedLocalDataType"
            ) -> model_types.SupportedLocalDataType:
                if not tensorflow_handler.SeqOfTensorflowTensorHandler.can_handle(sample_input):
                    sample_input = tensorflow_handler.SeqOfTensorflowTensorHandler.convert_from_df(
                        model_signature._convert_local_data_to_df(sample_input)
                    )

                target_method = getattr(model, target_method_name, None)
                assert callable(target_method)
                for tensor in sample_input:
                    tensorflow.stop_gradient(tensor)
                predictions_df = target_method(*sample_input)

                if isinstance(predictions_df, (tensorflow.Tensor, tensorflow.Variable, np.ndarray)):
                    predictions_df = [predictions_df]

                return predictions_df

            model_meta = handlers_utils.validate_signature(
                model=model,
                model_meta=model_meta,
                target_methods=target_methods,
                sample_input=sample_input,
                get_prediction_fn=get_prediction,
            )

        model_blob_path = os.path.join(model_blobs_dir_path, name)
        os.makedirs(model_blob_path, exist_ok=True)
        if isinstance(model, tensorflow.keras.Model):
            tensorflow.keras.models.save_model(model, os.path.join(model_blob_path, cls.MODELE_BLOB_FILE_OR_DIR))
        else:
            tensorflow.saved_model.save(model, os.path.join(model_blob_path, cls.MODELE_BLOB_FILE_OR_DIR))

        base_meta = model_blob_meta.ModelBlobMeta(
            name=name,
            model_type=cls.HANDLER_TYPE,
            handler_version=cls.HANDLER_VERSION,
            path=cls.MODELE_BLOB_FILE_OR_DIR,
        )
        model_meta.models[name] = base_meta
        model_meta.min_snowpark_ml_version = cls._MIN_SNOWPARK_ML_VERSION

        model_meta.env.include_if_absent(
            [model_env.ModelDependency(requirement="tensorflow", pip_name="tensorflow")], check_local_version=True
        )
        model_meta.env.cuda_version = kwargs.get("cuda_version", model_env.DEFAULT_CUDA_VERSION)

    @classmethod
    def load_model(
        cls,
        name: str,
        model_meta: model_meta_api.ModelMetadata,
        model_blobs_dir_path: str,
        **kwargs: Unpack[model_types.ModelLoadOption],
    ) -> "tensorflow.Module":
        import tensorflow

        model_blob_path = os.path.join(model_blobs_dir_path, name)
        model_blobs_metadata = model_meta.models
        model_blob_metadata = model_blobs_metadata[name]
        model_blob_filename = model_blob_metadata.path
        m = tensorflow.keras.models.load_model(os.path.join(model_blob_path, model_blob_filename), compile=False)
        if isinstance(m, tensorflow.keras.Model):
            return m
        return cast(tensorflow.Module, m)

    @classmethod
    def convert_as_custom_model(
        cls,
        raw_model: "tensorflow.Module",
        model_meta: model_meta_api.ModelMetadata,
        **kwargs: Unpack[model_types.ModelLoadOption],
    ) -> custom_model.CustomModel:
        import tensorflow

        from snowflake.ml.model import custom_model

        def _create_custom_model(
            raw_model: "tensorflow.Module",
            model_meta: model_meta_api.ModelMetadata,
        ) -> Type[custom_model.CustomModel]:
            def fn_factory(
                raw_model: "tensorflow.Module",
                signature: model_signature.ModelSignature,
                target_method: str,
            ) -> Callable[[custom_model.CustomModel, pd.DataFrame], pd.DataFrame]:
                @custom_model.inference_api
                def fn(self: custom_model.CustomModel, X: pd.DataFrame) -> pd.DataFrame:
                    if X.isnull().any(axis=None):
                        raise ValueError("Tensor cannot handle null values.")

                    t = tensorflow_handler.SeqOfTensorflowTensorHandler.convert_from_df(X, signature.inputs)

                    for tensor in t:
                        tensorflow.stop_gradient(tensor)
                    res = getattr(raw_model, target_method)(*t)

                    if isinstance(res, (tensorflow.Tensor, tensorflow.Variable, np.ndarray)):
                        res = [res]

                    if isinstance(res, list) and len(res) > 0 and isinstance(res[0], np.ndarray):
                        # In case of running on CPU, it will return numpy array
                        df = numpy_handler.SeqOfNumpyArrayHandler.convert_to_df(res)
                    else:
                        df = tensorflow_handler.SeqOfTensorflowTensorHandler.convert_to_df(res)
                    return model_signature_utils.rename_pandas_df(df, signature.outputs)

                return fn

            type_method_dict = {}
            for target_method_name, sig in model_meta.signatures.items():
                type_method_dict[target_method_name] = fn_factory(raw_model, sig, target_method_name)

            _TensorFlowModel = type(
                "_TensorFlowModel",
                (custom_model.CustomModel,),
                type_method_dict,
            )

            return _TensorFlowModel

        _TensorFlowModel = _create_custom_model(raw_model, model_meta)
        tf_model = _TensorFlowModel(custom_model.ModelContext())

        return tf_model



#!/usr/bin/env python3
from typing import Iterable, Optional, Union

import pandas as pd
from sklearn import preprocessing

from snowflake import snowpark
from snowflake.ml._internal import telemetry
from snowflake.ml._internal.exceptions import error_codes, exceptions
from snowflake.ml.modeling.framework import base
from snowflake.snowpark import functions as F, types as T


class Binarizer(base.BaseTransformer):
    r"""Binarizes data (sets feature values to 0 or 1) according to the given threshold.

    Values must be of float type. Values greater than the threshold map to 1, while values less than or equal to
    the threshold map to 0. The default threshold of 0.0 maps only positive values to 1.

    For more details on what this transformer does, see [sklearn.preprocessing.Binarizer]
    (https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.Binarizer.html).

    Args:
        threshold: Feature values below or equal to this are replaced by 0, above it by 1. Default values is 0.0.
        input_cols: The name(s) of one or more columns in a DataFrame containing a feature to be binarized.
        output_cols: The name(s) of one or more columns in a DataFrame in which results will be stored. The number of
            columns specified must match the number of input columns.
        drop_input_cols: Remove input columns from output if set True. False by default.
    """

    def __init__(
        self,
        *,
        threshold: float = 0.0,
        input_cols: Optional[Union[str, Iterable[str]]] = None,
        output_cols: Optional[Union[str, Iterable[str]]] = None,
        drop_input_cols: Optional[bool] = False,
    ) -> None:
        """
        Map each feature to a binary value.

        Values greater than the threshold map to 1, while values less than or equal to
        the threshold map to 0. With the default threshold of 0, only positive values map to 1.

        Data must be float-valued.

        Args:
            threshold: Feature values below or equal to this are replaced by 0, above it by 1. Default values is 0.0.
            input_cols: The name(s) of one or more columns in a DataFrame containing a feature to be binarized.
            output_cols: The name(s) of one or more columns in a DataFrame in which results will be stored. The number
                of columns specified must match the number of input columns.
            drop_input_cols: Remove input columns from output if set True. False by default.
        """
        super().__init__(drop_input_cols=drop_input_cols)
        self.threshold = threshold
        self.set_input_cols(input_cols)
        self.set_output_cols(output_cols)

    def _reset(self) -> None:
        """
        Reset internal data-dependent state of the binarizer.
        __init__ parameters are not touched.

        This is a stateless transformer, so there is nothing to reset.
        """
        super()._reset()

    @telemetry.send_api_usage_telemetry(
        project=base.PROJECT,
        subproject=base.SUBPROJECT,
    )
    def fit(self, dataset: Union[snowpark.DataFrame, pd.DataFrame]) -> "Binarizer":
        """
        This is a stateless transformer, so there is nothing to fit. Validates the transformer arguments.
        Returns the transformer instance.

        Args:
            dataset: Input dataset.

        Returns:
            self

        Raises:
            SnowflakeMLException: If the threshold is not a float.
        """
        if not isinstance(self.threshold, float):
            raise exceptions.SnowflakeMLException(
                error_code=error_codes.INVALID_ATTRIBUTE,
                original_exception=TypeError(f"Binarizer threshold must be a float, but got {type(self.threshold)}."),
            )

        self._is_fitted = True
        return self

    @telemetry.send_api_usage_telemetry(
        project=base.PROJECT,
        subproject=base.SUBPROJECT,
    )
    @telemetry.add_stmt_params_to_df(
        project=base.PROJECT,
        subproject=base.SUBPROJECT,
    )
    def transform(self, dataset: Union[snowpark.DataFrame, pd.DataFrame]) -> Union[snowpark.DataFrame, pd.DataFrame]:
        """
        Binarize the data. Map to 1 if it is strictly greater than the threshold, otherwise 0.

        Args:
            dataset: Input dataset.

        Returns:
            Output dataset.
        """
        super()._check_input_cols()
        super()._check_output_cols()
        super()._check_dataset_type(dataset)

        if isinstance(dataset, snowpark.DataFrame):
            output_df = self._transform_snowpark(dataset)
        else:
            output_df = self._transform_sklearn(dataset)

        return self._drop_input_columns(output_df) if self._drop_input_cols is True else output_df

    def _transform_snowpark(self, dataset: snowpark.DataFrame) -> snowpark.DataFrame:
        passthrough_columns = [c for c in dataset.columns if c not in self.output_cols]
        self._validate_data_has_no_nulls(dataset)
        output_columns = []
        for input_col in self.input_cols:
            col = F.iff(dataset[input_col] > self.threshold, 1.0, 0.0).cast(T.FloatType())
            output_columns.append(col)

        transformed_dataset: snowpark.DataFrame = dataset.with_columns(self.output_cols, output_columns)
        # Reorder columns. Passthrough columns are added at the right to the output of the transformers.
        transformed_dataset = transformed_dataset[self.output_cols + passthrough_columns]
        return transformed_dataset

    def _create_unfitted_sklearn_object(self) -> preprocessing.Binarizer:
        return preprocessing.Binarizer(threshold=self.threshold)

    def _create_sklearn_object(self) -> preprocessing.Binarizer:
        """
        Get an equivalent sklearn Binarizer.

        Returns:
            Sklearn Binarizer.
        """
        return self._create_unfitted_sklearn_object()



import importlib
import inspect
import io
import os
import posixpath
import sys
from typing import Any, Callable, Dict, List, Optional, Tuple, Union
from uuid import uuid4

import cloudpickle as cp
import numpy as np
import pandas as pd
import sklearn
from scipy.stats import rankdata
from sklearn import model_selection

from snowflake.ml._internal import telemetry
from snowflake.ml._internal.env_utils import SNOWML_SPROC_ENV
from snowflake.ml._internal.exceptions import (
    error_codes,
    exceptions,
    modeling_error_messages,
)
from snowflake.ml._internal.utils import identifier, snowpark_dataframe_utils
from snowflake.ml._internal.utils.query_result_checker import SqlResultValidator
from snowflake.ml._internal.utils.temp_file_utils import (
    cleanup_temp_files,
    get_temp_file_path,
)
from snowflake.snowpark import (
    DataFrame,
    Session,
    exceptions as snowpark_exceptions,
    functions as F,
)
from snowflake.snowpark._internal.utils import (
    TempObjectType,
    random_name_for_temp_object,
)
from snowflake.snowpark.functions import col, pandas_udf, sproc, udtf
from snowflake.snowpark.stored_procedure import StoredProcedure
from snowflake.snowpark.types import (
    IntegerType,
    PandasSeries,
    StringType,
    StructField,
    StructType,
)

cp.register_pickle_by_value(inspect.getmodule(get_temp_file_path))
cp.register_pickle_by_value(inspect.getmodule(identifier.get_inferred_name))

_PROJECT = "ModelDevelopment"


class WrapperProvider:
    def __init__(self) -> None:
        self.imports: List[str] = []
        self.dependencies: List[str] = []

    def get_fit_wrapper_function(
        self,
    ) -> Callable[[Any, List[str], str, str, List[str], List[str], Optional[str], Dict[str, str]], str]:
        imports = self.imports  # In order for the sproc to not resolve this reference in snowflake.ml

        def fit_wrapper_function(
            session: Session,
            sql_queries: List[str],
            stage_transform_file_name: str,
            stage_result_file_name: str,
            input_cols: List[str],
            label_cols: List[str],
            sample_weight_col: Optional[str],
            statement_params: Dict[str, str],
        ) -> str:
            import inspect
            import os

            import cloudpickle as cp
            import pandas as pd

            for import_name in imports:
                importlib.import_module(import_name)

            # Execute snowpark queries and obtain the results as pandas dataframe
            # NB: this implies that the result data must fit into memory.
            for query in sql_queries[:-1]:
                _ = session.sql(query).collect(statement_params=statement_params)
            sp_df = session.sql(sql_queries[-1])
            df: pd.DataFrame = sp_df.to_pandas(statement_params=statement_params)
            df.columns = sp_df.columns

            local_transform_file_name = get_temp_file_path()

            session.file.get(stage_transform_file_name, local_transform_file_name, statement_params=statement_params)

            local_transform_file_path = os.path.join(
                local_transform_file_name, os.listdir(local_transform_file_name)[0]
            )
            with open(local_transform_file_path, mode="r+b") as local_transform_file_obj:
                estimator = cp.load(local_transform_file_obj)

            argspec = inspect.getfullargspec(estimator.fit)
            args = {"X": df[input_cols]}
            if label_cols:
                label_arg_name = "Y" if "Y" in argspec.args else "y"
                args[label_arg_name] = df[label_cols].squeeze()

            if sample_weight_col is not None and "sample_weight" in argspec.args:
                args["sample_weight"] = df[sample_weight_col].squeeze()

            estimator.fit(**args)

            local_result_file_name = get_temp_file_path()

            with open(local_result_file_name, mode="w+b") as local_result_file_obj:
                cp.dump(estimator, local_result_file_obj)

            session.file.put(
                local_result_file_name,
                stage_result_file_name,
                auto_compress=False,
                overwrite=True,
                statement_params=statement_params,
            )

            # Note: you can add something like  + "|" + str(df) to the return string
            # to pass debug information to the caller.
            return str(os.path.basename(local_result_file_name))

        return fit_wrapper_function


class SklearnWrapperProvider(WrapperProvider):
    def __init__(self) -> None:
        import sklearn

        self.imports: List[str] = ["sklearn"]

        # TODO(snandamuri): Replace cloudpickle with joblib after latest version of joblib is added to snowflake conda.
        self.dependencies: List[str] = [
            f"numpy=={np.__version__}",
            f"scikit-learn=={sklearn.__version__}",
            f"cloudpickle=={cp.__version__}",
        ]


class XGBoostWrapperProvider(WrapperProvider):
    def __init__(self) -> None:
        import xgboost

        self.imports: List[str] = ["xgboost"]
        self.dependencies = [
            f"numpy=={np.__version__}",
            f"xgboost=={xgboost.__version__}",
            f"cloudpickle=={cp.__version__}",
        ]


class LightGBMWrapperProvider(WrapperProvider):
    def __init__(self) -> None:
        import lightgbm

        self.imports: List[str] = ["lightgbm"]
        self.dependencies = [
            f"numpy=={np.__version__}",
            f"lightgbm=={lightgbm.__version__}",
            f"cloudpickle=={cp.__version__}",
        ]


class SklearnModelSelectionWrapperProvider(WrapperProvider):
    def __init__(self) -> None:
        import xgboost

        self.imports: List[str] = ["sklearn", "xgboost"]
        self.dependencies = [
            f"numpy=={np.__version__}",
            f"scikit-learn=={sklearn.__version__}",
            f"cloudpickle=={cp.__version__}",
            f"xgboost=={xgboost.__version__}",
        ]

        # Only include lightgbm in the dependencies if it is installed.
        try:
            import lightgbm
        except ModuleNotFoundError:
            pass
        else:
            self.imports.append("lightgbm")
            self.dependencies.append(f"lightgbm=={lightgbm.__version__}")


def _get_rand_id() -> str:
    """
    Generate random id to be used in sproc and stage names.

    Returns:
        Random id string usable in sproc, table, and stage names.
    """
    return str(uuid4()).replace("-", "_").upper()


class SnowparkHandlers:
    def __init__(
        self, class_name: str, subproject: str, wrapper_provider: WrapperProvider, autogenerated: Optional[bool] = False
    ) -> None:
        self._class_name = class_name
        self._subproject = subproject
        self._wrapper_provider = wrapper_provider
        self._autogenerated = autogenerated

    def _get_fit_wrapper_sproc(
        self, dependencies: List[str], session: Session, statement_params: Dict[str, str]
    ) -> StoredProcedure:
        # If the sproc already exists, don't register.
        if not hasattr(session, "_FIT_WRAPPER_SPROCS"):
            session._FIT_WRAPPER_SPROCS: Dict[str, StoredProcedure] = {}  # type: ignore[attr-defined, misc]

        fit_sproc_key = self._wrapper_provider.__class__.__name__
        if fit_sproc_key in session._FIT_WRAPPER_SPROCS:  # type: ignore[attr-defined]
            fit_sproc: StoredProcedure = session._FIT_WRAPPER_SPROCS[fit_sproc_key]  # type: ignore[attr-defined]
            return fit_sproc

        fit_sproc_name = random_name_for_temp_object(TempObjectType.PROCEDURE)

        fit_wrapper_sproc = session.sproc.register(
            func=self._wrapper_provider.get_fit_wrapper_function(),
            is_permanent=False,
            name=fit_sproc_name,
            packages=dependencies,  # type: ignore[arg-type]
            replace=True,
            session=session,
            statement_params=statement_params,
        )

        session._FIT_WRAPPER_SPROCS[fit_sproc_key] = fit_wrapper_sproc  # type: ignore[attr-defined]

        return fit_wrapper_sproc

    def fit_pandas(
        self,
        dataset: pd.DataFrame,
        estimator: object,
        input_cols: List[str],
        label_cols: Optional[List[str]],
        sample_weight_col: Optional[str],
    ) -> object:
        assert hasattr(estimator, "fit")  # Keep mypy happy
        argspec = inspect.getfullargspec(estimator.fit)
        args = {"X": dataset[input_cols]}

        if label_cols:
            label_arg_name = "Y" if "Y" in argspec.args else "y"
            args[label_arg_name] = dataset[label_cols].squeeze()

        if sample_weight_col is not None and "sample_weight" in argspec.args:
            args["sample_weight"] = dataset[sample_weight_col].squeeze()

        return estimator.fit(**args)

    def fit_snowpark(
        self,
        dataset: DataFrame,
        session: Session,
        estimator: object,
        dependencies: List[str],
        input_cols: List[str],
        label_cols: List[str],
        sample_weight_col: Optional[str],
    ) -> Any:
        dataset = snowpark_dataframe_utils.cast_snowpark_dataframe_column_types(dataset)

        # If we are already in a stored procedure, no need to kick off another one.
        if SNOWML_SPROC_ENV in os.environ:
            statement_params = telemetry.get_function_usage_statement_params(
                project=_PROJECT,
                subproject=self._subproject,
                function_name=telemetry.get_statement_params_full_func_name(inspect.currentframe(), self._class_name),
                api_calls=[Session.call],
                custom_tags=dict([("autogen", True)]) if self._autogenerated else None,
            )
            pd_df: pd.DataFrame = dataset.to_pandas(statement_params=statement_params)
            pd_df.columns = dataset.columns
            return self.fit_pandas(pd_df, estimator, input_cols, label_cols, sample_weight_col)

        # Extract query that generated the dataframe. We will need to pass it to the fit procedure.
        queries = dataset.queries["queries"]

        # Create a temp file and dump the transform to that file.
        local_transform_file_name = get_temp_file_path()
        with open(local_transform_file_name, mode="w+b") as local_transform_file:
            cp.dump(estimator, local_transform_file)

        # Create temp stage to run fit.
        transform_stage_name = random_name_for_temp_object(TempObjectType.STAGE)
        stage_creation_query = f"CREATE OR REPLACE TEMPORARY STAGE {transform_stage_name};"
        SqlResultValidator(session=session, query=stage_creation_query).has_dimensions(
            expected_rows=1, expected_cols=1
        ).validate()

        # Use posixpath to construct stage paths
        stage_transform_file_name = posixpath.join(transform_stage_name, os.path.basename(local_transform_file_name))
        stage_result_file_name = posixpath.join(transform_stage_name, os.path.basename(local_transform_file_name))
        local_result_file_name = get_temp_file_path()

        statement_params = telemetry.get_function_usage_statement_params(
            project=_PROJECT,
            subproject=self._subproject,
            function_name=telemetry.get_statement_params_full_func_name(inspect.currentframe(), self._class_name),
            api_calls=[sproc],
            custom_tags=dict([("autogen", True)]) if self._autogenerated else None,
        )
        # Put locally serialized transform on stage.
        session.file.put(
            local_transform_file_name,
            stage_transform_file_name,
            auto_compress=False,
            overwrite=True,
            statement_params=statement_params,
        )

        # Call fit sproc
        statement_params = telemetry.get_function_usage_statement_params(
            project=_PROJECT,
            subproject=self._subproject,
            function_name=telemetry.get_statement_params_full_func_name(inspect.currentframe(), self._class_name),
            api_calls=[Session.call],
            custom_tags=dict([("autogen", True)]) if self._autogenerated else None,
        )

        fit_wrapper_sproc = self._get_fit_wrapper_sproc(dependencies, session, statement_params)

        try:
            sproc_export_file_name: str = fit_wrapper_sproc(
                session,
                queries,
                stage_transform_file_name,
                stage_result_file_name,
                input_cols,
                label_cols,
                sample_weight_col,
                statement_params,
            )
        except snowpark_exceptions.SnowparkClientException as e:
            if "fit() missing 1 required positional argument: 'y'" in str(e):
                raise exceptions.SnowflakeMLException(
                    error_code=error_codes.NOT_FOUND,
                    original_exception=RuntimeError(modeling_error_messages.ATTRIBUTE_NOT_SET.format("label_cols")),
                ) from e
            raise e

        if "|" in sproc_export_file_name:
            fields = sproc_export_file_name.strip().split("|")
            sproc_export_file_name = fields[0]

        session.file.get(
            posixpath.join(stage_result_file_name, sproc_export_file_name),
            local_result_file_name,
            statement_params=statement_params,
        )

        with open(os.path.join(local_result_file_name, sproc_export_file_name), mode="r+b") as result_file_obj:
            fit_estimator = cp.load(result_file_obj)

        cleanup_temp_files([local_transform_file_name, local_result_file_name])

        return fit_estimator

    def batch_inference(
        self,
        dataset: DataFrame,
        session: Session,
        estimator: object,
        dependencies: List[str],
        inference_method: str,
        input_cols: List[str],
        pass_through_columns: List[str],
        expected_output_cols_list: List[str],
        expected_output_cols_type: str = "",
        *args: Any,
        **kwargs: Any,
    ) -> DataFrame:
        # Register vectorized UDF for batch inference
        batch_inference_udf_name = random_name_for_temp_object(TempObjectType.FUNCTION)
        snowpark_cols = dataset.select(input_cols).columns
        dataset = snowpark_dataframe_utils.cast_snowpark_dataframe_column_types(dataset)

        statement_params = telemetry.get_function_usage_statement_params(
            project=_PROJECT,
            subproject=self._subproject,
            function_name=telemetry.get_statement_params_full_func_name(inspect.currentframe(), self._class_name),
            api_calls=[pandas_udf],
            custom_tags=dict([("autogen", True)]) if self._autogenerated else None,
        )

        @pandas_udf(  # type: ignore[arg-type, misc]
            is_permanent=False,
            name=batch_inference_udf_name,
            packages=dependencies,  # type: ignore[arg-type]
            replace=True,
            session=session,
            statement_params=statement_params,
        )
        def vec_batch_infer(ds: PandasSeries[dict]) -> PandasSeries[dict]:  # type: ignore[type-arg]
            import numpy as np
            import pandas as pd

            input_df = pd.json_normalize(ds)

            # pd.json_normalize() doesn't remove quotes around quoted identifiers like snowpakr_df.to_pandas().
            # But trained models have unquoted input column names saved in internal state if trained using snowpark_df
            # or quoted input column names saved in internal state if trained using pandas_df.
            # Model expects exact same columns names in the input df for predict call.

            input_df = input_df[input_cols]  # Select input columns with quoted column names.
            if hasattr(estimator, "feature_names_in_"):
                missing_features = []
                for i, f in enumerate(getattr(estimator, "feature_names_in_", {})):
                    if i >= len(input_cols) or (input_cols[i] != f and snowpark_cols[i] != f):
                        missing_features.append(f)

                if len(missing_features) > 0:
                    raise ValueError(
                        "The feature names should match with those that were passed during fit.\n"
                        f"Features seen during fit call but not present in the input: {missing_features}\n"
                        f"Features in the input dataframe : {input_cols}\n"
                    )
                input_df.columns = getattr(estimator, "feature_names_in_", {})
            else:
                # Just rename the column names to unquoted identifiers.
                input_df.columns = snowpark_cols  # Replace the quoted columns identifier with unquoted column ids.
            inference_res = getattr(estimator, inference_method)(input_df, *args, **kwargs)
            if isinstance(inference_res, list) and len(inference_res) > 0 and isinstance(inference_res[0], np.ndarray):
                # In case of multioutput estimators, predict_proba, decision_function etc., functions return a list of
                # ndarrays. We need to concatenate them.
                transformed_numpy_array = np.concatenate(inference_res, axis=1)
            elif (
                isinstance(inference_res, tuple) and len(inference_res) > 0 and isinstance(inference_res[0], np.ndarray)
            ):
                # In case of kneighbors, functions return a tuple of ndarrays.
                transformed_numpy_array = np.stack(inference_res, axis=1)
            else:
                transformed_numpy_array = inference_res

            if (len(transformed_numpy_array.shape) == 3) and inference_method != "kneighbors":
                # VotingClassifier will return results of shape (n_classifiers, n_samples, n_classes)
                # when voting = "soft" and flatten_transform = False. We can't handle unflatten transforms,
                # so we ignore flatten_transform flag and flatten the results.
                transformed_numpy_array = np.hstack(transformed_numpy_array)  # type: ignore[call-overload]

            if len(transformed_numpy_array.shape) > 1:
                if transformed_numpy_array.shape[1] != len(expected_output_cols_list):
                    # HeterogeneousEnsemble's transform method produce results with variying shapes
                    # from (n_samples, n_estimators) to (n_samples, n_estimators * n_classes).
                    # It is hard to predict the response shape without using fragile introspection logic.
                    # So, to avoid that we are packing the results into a dataframe of shape (n_samples, 1) with
                    # each element being a list.
                    if len(expected_output_cols_list) != 1:
                        raise TypeError(
                            "expected_output_cols_list must be same length as transformed array or "
                            "should be of length 1"
                        )
                    series = pd.Series(transformed_numpy_array.tolist())
                    transformed_pandas_df = pd.DataFrame(series, columns=expected_output_cols_list)
                else:
                    transformed_pandas_df = pd.DataFrame(
                        transformed_numpy_array.tolist(), columns=expected_output_cols_list
                    )
            else:
                transformed_pandas_df = pd.DataFrame(transformed_numpy_array, columns=expected_output_cols_list)

            return transformed_pandas_df.to_dict("records")  # type: ignore[no-any-return]

        batch_inference_table_name = f"SNOWML_BATCH_INFERENCE_INPUT_TABLE_{_get_rand_id()}"

        # Run Transform
        query_from_df = str(dataset.queries["queries"][0])

        outer_select_list = pass_through_columns[:]
        inner_select_list = pass_through_columns[:]

        outer_select_list.extend(
            [
                "{object_name}:{column_name}{udf_datatype} as {column_name}".format(
                    object_name=batch_inference_udf_name,
                    column_name=identifier.get_inferred_name(c),
                    udf_datatype=(f"::{expected_output_cols_type}" if expected_output_cols_type else ""),
                )
                for c in expected_output_cols_list
            ]
        )

        inner_select_list.extend(
            [
                "{udf_name}(object_construct_keep_null({input_cols_dict})) AS {udf_name}".format(
                    udf_name=batch_inference_udf_name,
                    input_cols_dict=", ".join([f"'{c}', {c}" for c in input_cols]),
                )
            ]
        )

        sql = """WITH {input_table_name} AS ({query})
                    SELECT
                      {outer_select_stmt}
                    FROM (
                      SELECT
                        {inner_select_stmt}
                      FROM {input_table_name}
                    )
               """.format(
            input_table_name=batch_inference_table_name,
            query=query_from_df,
            outer_select_stmt=", ".join(outer_select_list),
            inner_select_stmt=", ".join(inner_select_list),
        )

        return session.sql(sql)

    def score_pandas(
        self,
        dataset: pd.DataFrame,
        estimator: object,
        input_cols: List[str],
        label_cols: List[str],
        sample_weight_col: Optional[str],
    ) -> float:
        assert hasattr(estimator, "score")  # make type checker happy
        argspec = inspect.getfullargspec(estimator.score)
        if "X" in argspec.args:
            args = {"X": dataset[input_cols]}
        elif "X_test" in argspec.args:
            args = {"X_test": dataset[input_cols]}
        else:
            raise exceptions.SnowflakeMLException(
                error_code=error_codes.INVALID_ATTRIBUTE,
                original_exception=RuntimeError("Neither 'X' or 'X_test' exist in argument"),
            )

        if len(label_cols) > 0:
            label_arg_name = "Y" if "Y" in argspec.args else "y"
            args[label_arg_name] = dataset[label_cols].squeeze()

        if sample_weight_col is not None and "sample_weight" in argspec.args:
            args["sample_weight"] = dataset[sample_weight_col].squeeze()

        score = estimator.score(**args)
        assert isinstance(score, float)  # make type checker happy

        return score

    def score_snowpark(
        self,
        dataset: DataFrame,
        session: Session,
        estimator: object,
        dependencies: List[str],
        score_sproc_imports: List[str],
        input_cols: List[str],
        label_cols: List[str],
        sample_weight_col: Optional[str],
    ) -> float:
        dataset = snowpark_dataframe_utils.cast_snowpark_dataframe_column_types(dataset)
        if SNOWML_SPROC_ENV in os.environ:
            statement_params = telemetry.get_function_usage_statement_params(
                project=_PROJECT,
                subproject=self._subproject,
                function_name=telemetry.get_statement_params_full_func_name(inspect.currentframe(), self._class_name),
                api_calls=[Session.call],
                custom_tags=dict([("autogen", True)]) if self._autogenerated else None,
            )
            pd_df: pd.DataFrame = dataset.to_pandas(statement_params=statement_params)
            pd_df.columns = dataset.columns
            return self.score_pandas(pd_df, estimator, input_cols, label_cols, sample_weight_col)

        # Extract queries that generated the dataframe. We will need to pass it to score procedure.
        queries = dataset.queries["queries"]

        # Create a temp file and dump the score to that file.
        local_score_file_name = get_temp_file_path()
        with open(local_score_file_name, mode="w+b") as local_score_file:
            cp.dump(estimator, local_score_file)

        # Create temp stage to run score.
        score_stage_name = random_name_for_temp_object(TempObjectType.STAGE)
        assert session is not None  # keep mypy happy
        stage_creation_query = f"CREATE OR REPLACE TEMPORARY STAGE {score_stage_name};"
        SqlResultValidator(session=session, query=stage_creation_query).has_dimensions(
            expected_rows=1, expected_cols=1
        ).validate()

        # Use posixpath to construct stage paths
        stage_score_file_name = posixpath.join(score_stage_name, os.path.basename(local_score_file_name))
        score_sproc_name = random_name_for_temp_object(TempObjectType.PROCEDURE)
        statement_params = telemetry.get_function_usage_statement_params(
            project=_PROJECT,
            subproject=self._subproject,
            function_name=telemetry.get_statement_params_full_func_name(
                inspect.currentframe(), self.__class__.__name__
            ),
            api_calls=[sproc],
            custom_tags=dict([("autogen", True)]) if self._autogenerated else None,
        )
        # Put locally serialized score on stage.
        session.file.put(
            local_score_file_name,
            stage_score_file_name,
            auto_compress=False,
            overwrite=True,
            statement_params=statement_params,
        )

        @sproc(  # type: ignore[misc]
            is_permanent=False,
            name=score_sproc_name,
            packages=dependencies,  # type: ignore[arg-type]
            replace=True,
            session=session,
            statement_params=statement_params,
            anonymous=True,
        )
        def score_wrapper_sproc(
            session: Session,
            sql_queries: List[str],
            stage_score_file_name: str,
            input_cols: List[str],
            label_cols: List[str],
            sample_weight_col: Optional[str],
            statement_params: Dict[str, str],
        ) -> float:
            import inspect
            import os

            import cloudpickle as cp

            for import_name in score_sproc_imports:
                importlib.import_module(import_name)

            for query in sql_queries[:-1]:
                _ = session.sql(query).collect(statement_params=statement_params)
            sp_df = session.sql(sql_queries[-1])
            df: pd.DataFrame = sp_df.to_pandas(statement_params=statement_params)
            df.columns = sp_df.columns

            local_score_file_name = get_temp_file_path()
            session.file.get(stage_score_file_name, local_score_file_name, statement_params=statement_params)

            local_score_file_name_path = os.path.join(local_score_file_name, os.listdir(local_score_file_name)[0])
            with open(local_score_file_name_path, mode="r+b") as local_score_file_obj:
                estimator = cp.load(local_score_file_obj)

            argspec = inspect.getfullargspec(estimator.score)
            if "X" in argspec.args:
                args = {"X": df[input_cols]}
            elif "X_test" in argspec.args:
                args = {"X_test": df[input_cols]}
            else:
                raise RuntimeError("Neither 'X' or 'X_test' exist in argument")

            if label_cols:
                label_arg_name = "Y" if "Y" in argspec.args else "y"
                args[label_arg_name] = df[label_cols].squeeze()

            if sample_weight_col is not None and "sample_weight" in argspec.args:
                args["sample_weight"] = df[sample_weight_col].squeeze()

            result: float = estimator.score(**args)
            return result

        # Call score sproc
        statement_params = telemetry.get_function_usage_statement_params(
            project=_PROJECT,
            subproject=self._subproject,
            function_name=telemetry.get_statement_params_full_func_name(
                inspect.currentframe(), self.__class__.__name__
            ),
            api_calls=[Session.call],
            custom_tags=dict([("autogen", True)]) if self._autogenerated else None,
        )
        score: float = score_wrapper_sproc(
            session,
            queries,
            stage_score_file_name,
            input_cols,
            label_cols,
            sample_weight_col,
            statement_params,
        )

        cleanup_temp_files([local_score_file_name])

        return score

    def fit_search_snowpark(
        self,
        param_grid: Union[model_selection.ParameterGrid, model_selection.ParameterSampler],
        dataset: DataFrame,
        session: Session,
        estimator: Union[model_selection.GridSearchCV, model_selection.RandomizedSearchCV],
        dependencies: List[str],
        udf_imports: List[str],
        input_cols: List[str],
        label_cols: List[str],
        sample_weight_col: Optional[str],
    ) -> Union[model_selection.GridSearchCV, model_selection.RandomizedSearchCV]:
        from itertools import product

        import cachetools
        from sklearn.base import clone, is_classifier
        from sklearn.calibration import check_cv

        # Create one stage for data and for estimators.
        temp_stage_name = random_name_for_temp_object(TempObjectType.STAGE)
        temp_stage_creation_query = f"CREATE OR REPLACE TEMP STAGE {temp_stage_name};"
        session.sql(temp_stage_creation_query).collect()

        # Stage data.
        dataset = snowpark_dataframe_utils.cast_snowpark_dataframe(dataset)
        remote_file_path = f"{temp_stage_name}/{temp_stage_name}.parquet"
        dataset.write.copy_into_location(  # type:ignore[call-overload]
            remote_file_path, file_format_type="parquet", header=True, overwrite=True
        )
        imports = [f"@{row.name}" for row in session.sql(f"LIST @{temp_stage_name}").collect()]

        # Store GridSearchCV's refit variable. If user set it as False, we don't need to refit it again
        original_refit = estimator.refit

        # Create a temp file and dump the estimator to that file.
        estimator_file_name = get_temp_file_path()
        params_to_evaluate = []
        for param_to_eval in list(param_grid):
            for k, v in param_to_eval.items():
                param_to_eval[k] = [v]
            params_to_evaluate.append([param_to_eval])

        with open(estimator_file_name, mode="w+b") as local_estimator_file_obj:
            # Set GridSearchCV refit as False and fit it again after retrieving the best param
            estimator.refit = False
            cp.dump(dict(estimator=estimator, param_grid=params_to_evaluate), local_estimator_file_obj)
        stage_estimator_file_name = posixpath.join(temp_stage_name, os.path.basename(estimator_file_name))
        sproc_statement_params = telemetry.get_function_usage_statement_params(
            project=_PROJECT,
            subproject=self._subproject,
            function_name=telemetry.get_statement_params_full_func_name(
                inspect.currentframe(), self.__class__.__name__
            ),
            api_calls=[sproc],
            custom_tags=dict([("autogen", True)]) if self._autogenerated else None,
        )
        udtf_statement_params = telemetry.get_function_usage_statement_params(
            project=_PROJECT,
            subproject=self._subproject,
            function_name=telemetry.get_statement_params_full_func_name(
                inspect.currentframe(), self.__class__.__name__
            ),
            api_calls=[udtf],
            custom_tags=dict([("autogen", True)]) if self._autogenerated else None,
        )

        # Put locally serialized estimator on stage.
        put_result = session.file.put(
            estimator_file_name,
            temp_stage_name,
            auto_compress=False,
            overwrite=True,
        )
        estimator_location = put_result[0].target
        imports.append(f"@{temp_stage_name}/{estimator_location}")

        search_sproc_name = random_name_for_temp_object(TempObjectType.PROCEDURE)
        random_udtf_name = random_name_for_temp_object(TempObjectType.FUNCTION)

        required_deps = dependencies + [
            "snowflake-snowpark-python<2",
            "fastparquet<2023.11",
            "pyarrow<14",
            "cachetools<5",
        ]

        @sproc(  # type: ignore[misc]
            is_permanent=False,
            name=search_sproc_name,
            packages=required_deps,  # type: ignore[arg-type]
            replace=True,
            session=session,
            anonymous=True,
            imports=imports,  # type: ignore[arg-type]
            statement_params=sproc_statement_params,
        )
        def _distributed_search(
            session: Session,
            imports: List[str],
            stage_estimator_file_name: str,
            input_cols: List[str],
            label_cols: List[str],
        ) -> str:
            import os
            import time
            from typing import Iterator

            import cloudpickle as cp
            import pandas as pd
            import pyarrow.parquet as pq
            from sklearn.metrics import check_scoring
            from sklearn.metrics._scorer import _check_multimetric_scoring

            for import_name in udf_imports:
                importlib.import_module(import_name)

            data_files = [
                filename
                for filename in os.listdir(sys._xoptions["snowflake_import_directory"])
                if filename.startswith(temp_stage_name)
            ]
            partial_df = [
                pq.read_table(os.path.join(sys._xoptions["snowflake_import_directory"], file_name)).to_pandas()
                for file_name in data_files
            ]
            df = pd.concat(partial_df, ignore_index=True)
            df.columns = [identifier.get_inferred_name(col) for col in df.columns]

            X = df[input_cols]
            y = df[label_cols].squeeze()

            local_estimator_file_name = get_temp_file_path()
            session.file.get(stage_estimator_file_name, local_estimator_file_name)

            local_estimator_file_path = os.path.join(
                local_estimator_file_name, os.listdir(local_estimator_file_name)[0]
            )
            with open(local_estimator_file_path, mode="r+b") as local_estimator_file_obj:
                estimator = cp.load(local_estimator_file_obj)["estimator"]

            cv_orig = check_cv(estimator.cv, y, classifier=is_classifier(estimator.estimator))
            indices = [test for _, test in cv_orig.split(X, y)]
            local_indices_file_name = get_temp_file_path()
            with open(local_indices_file_name, mode="w+b") as local_indices_file_obj:
                cp.dump(indices, local_indices_file_obj)

            # Put locally serialized indices on stage.
            put_result = session.file.put(
                local_indices_file_name,
                temp_stage_name,
                auto_compress=False,
                overwrite=True,
            )
            indices_location = put_result[0].target
            imports.append(f"@{temp_stage_name}/{indices_location}")
            indices_len = len(indices)

            assert estimator is not None

            @cachetools.cached(cache={})
            def _load_data_into_udf() -> Tuple[
                Dict[str, pd.DataFrame],
                Union[model_selection.GridSearchCV, model_selection.RandomizedSearchCV],
                pd.DataFrame,
                int,
                List[Dict[str, Any]],
            ]:
                import pyarrow.parquet as pq

                data_files = [
                    filename
                    for filename in os.listdir(sys._xoptions["snowflake_import_directory"])
                    if filename.startswith(temp_stage_name)
                ]
                partial_df = [
                    pq.read_table(os.path.join(sys._xoptions["snowflake_import_directory"], file_name)).to_pandas()
                    for file_name in data_files
                ]
                df = pd.concat(partial_df, ignore_index=True)
                df.columns = [identifier.get_inferred_name(col) for col in df.columns]

                # load estimator
                local_estimator_file_path = os.path.join(
                    sys._xoptions["snowflake_import_directory"], f"{estimator_location}"
                )
                with open(local_estimator_file_path, mode="rb") as local_estimator_file_obj:
                    estimator_objects = cp.load(local_estimator_file_obj)
                    estimator = estimator_objects["estimator"]
                    params_to_evaluate = estimator_objects["param_grid"]

                # load indices
                local_indices_file_path = os.path.join(
                    sys._xoptions["snowflake_import_directory"], f"{indices_location}"
                )
                with open(local_indices_file_path, mode="rb") as local_indices_file_obj:
                    indices = cp.load(local_indices_file_obj)

                argspec = inspect.getfullargspec(estimator.fit)
                args = {"X": df[input_cols]}

                if label_cols:
                    label_arg_name = "Y" if "Y" in argspec.args else "y"
                    args[label_arg_name] = df[label_cols].squeeze()

                if sample_weight_col is not None and "sample_weight" in argspec.args:
                    args["sample_weight"] = df[sample_weight_col].squeeze()
                return args, estimator, indices, len(df), params_to_evaluate

            class SearchCV:
                def __init__(self) -> None:
                    args, estimator, indices, data_length, params_to_evaluate = _load_data_into_udf()
                    self.args = args
                    self.estimator = estimator
                    self.indices = indices
                    self.data_length = data_length
                    self.params_to_evaluate = params_to_evaluate

                def process(self, params_idx: int, idx: int) -> Iterator[Tuple[str]]:
                    if hasattr(estimator, "param_grid"):
                        self.estimator.param_grid = self.params_to_evaluate[params_idx]
                    else:
                        self.estimator.param_distributions = self.params_to_evaluate[params_idx]
                    full_indices = np.array([i for i in range(self.data_length)])
                    test_indice = self.indices[idx]
                    train_indice = np.setdiff1d(full_indices, test_indice)
                    self.estimator.cv = [(train_indice, test_indice)]
                    self.estimator.fit(**self.args)
                    binary_cv_results = None
                    with io.BytesIO() as f:
                        cp.dump(self.estimator.cv_results_, f)
                        f.seek(0)
                        binary_cv_results = f.getvalue().hex()
                    yield (binary_cv_results,)

                def end_partition(self) -> None:
                    ...

            session.udtf.register(
                SearchCV,
                output_schema=StructType([StructField("CV_RESULTS", StringType())]),
                input_types=[IntegerType(), IntegerType()],
                name=random_udtf_name,
                packages=required_deps,  # type: ignore[arg-type]
                replace=True,
                is_permanent=False,
                imports=imports,  # type: ignore[arg-type]
                statement_params=udtf_statement_params,
            )

            HP_TUNING = F.table_function(random_udtf_name)

            idx_length = int(indices_len)
            params_length = len(param_grid)
            idxs = [i for i in range(idx_length)]
            param_indices, training_indices = [], []
            for param_idx, cv_idx in product([param_index for param_index in range(params_length)], idxs):
                param_indices.append(param_idx)
                training_indices.append(cv_idx)

            pd_df = pd.DataFrame(
                {
                    "PARAMS": param_indices,
                    "TRAIN_IND": training_indices,
                    "PARAM_INDEX": [i for i in range(idx_length * params_length)],
                }
            )
            df = session.create_dataframe(pd_df)
            results = df.select(
                F.cast(df["PARAM_INDEX"], IntegerType()).as_("PARAM_INDEX"),
                (HP_TUNING(df["PARAMS"], df["TRAIN_IND"]).over(partition_by=df["PARAM_INDEX"])),
            )

            # cv_result maintains the original order
            multimetric = False
            cv_results_ = dict()
            scorers = set()
            for i, val in enumerate(results.select("CV_RESULTS").sort(col("PARAM_INDEX")).collect()):
                # retrieved string had one more double quote in the front and end of the string.
                # use [1:-1] to remove the extra double quotes
                hex_str = bytes.fromhex(val[0])
                with io.BytesIO(hex_str) as f_reload:
                    each_cv_result = cp.load(f_reload)
                    for k, v in each_cv_result.items():
                        cur_cv = i % idx_length
                        key = k
                        if "split0_test_" in k:
                            # For multi-metric evaluation, the scores for all the scorers are available in the
                            # cv_results_ dict at the keys ending with that scorer’s name ('_<scorer_name>')
                            # instead of '_score'.
                            scorers.add(k[len("split0_test_") :])
                            key = k.replace("split0_test", f"split{cur_cv}_test")
                        elif k.startswith("param"):
                            if cur_cv != 0:
                                key = False
                        if key:
                            if key not in cv_results_:
                                cv_results_[key] = v
                            else:
                                cv_results_[key] = np.concatenate([cv_results_[key], v])

            multimetric = len(scorers) > 1
            # Use numpy to re-calculate all the information in cv_results_ again
            # Generally speaking, reshape all the results into the (scorers+2, idx_length, params_length) shape,
            # and average them by the idx_length;
            # idx_length is the number of cv folds; params_length is the number of parameter combinations
            scores = [
                np.reshape(
                    np.concatenate([cv_results_[f"split{cur_cv}_test_{score}"] for cur_cv in range(idx_length)]),
                    (idx_length, -1),
                )
                for score in scorers
            ]

            fit_score_test_matrix = np.stack(
                [
                    np.reshape(cv_results_["mean_fit_time"], (idx_length, -1)),
                    np.reshape(cv_results_["mean_score_time"], (idx_length, -1)),
                ]
                + scores
            )

            mean_fit_score_test_matrix = np.mean(fit_score_test_matrix, axis=1)
            std_fit_score_test_matrix = np.std(fit_score_test_matrix, axis=1)
            cv_results_["std_fit_time"] = std_fit_score_test_matrix[0]
            cv_results_["mean_fit_time"] = mean_fit_score_test_matrix[0]
            cv_results_["std_score_time"] = std_fit_score_test_matrix[1]
            cv_results_["mean_score_time"] = mean_fit_score_test_matrix[1]
            for idx, score in enumerate(scorers):
                cv_results_[f"std_test_{score}"] = std_fit_score_test_matrix[idx + 2]
                cv_results_[f"mean_test_{score}"] = mean_fit_score_test_matrix[idx + 2]
                # re-compute the ranking again with mean_test_<score>.
                cv_results_[f"rank_test_{score}"] = rankdata(-cv_results_[f"mean_test_{score}"], method="min")
                # The best param is the highest ranking (which is 1) and we choose the first time ranking 1 appeared.
                # If all scores are `nan`, `rankdata` will also produce an array of `nan` values.
                # In that case, default to first index.
                best_param_index = (
                    np.where(cv_results_[f"rank_test_{score}"] == 1)[0][0]
                    if not np.isnan(cv_results_[f"rank_test_{score}"]).all()
                    else 0
                )

            estimator.cv_results_ = cv_results_
            estimator.multimetric_ = multimetric

            # Reconstruct the sklearn estimator.
            refit_metric = "score"
            if callable(estimator.scoring):
                scorers = estimator.scoring
            elif estimator.scoring is None or isinstance(estimator.scoring, str):
                scorers = check_scoring(estimator.estimator, estimator.scoring)
            else:
                scorers = _check_multimetric_scoring(estimator.estimator, estimator.scoring)
                estimator._check_refit_for_multimetric(scorers)
                refit_metric = original_refit

            estimator.scorer_ = scorers

            # check refit_metric now for a callabe scorer that is multimetric
            if callable(estimator.scoring) and estimator.multimetric_:
                refit_metric = original_refit

            # For multi-metric evaluation, store the best_index_, best_params_ and
            # best_score_ iff refit is one of the scorer names
            # In single metric evaluation, refit_metric is "score"
            if original_refit or not estimator.multimetric_:
                estimator.best_index_ = estimator._select_best_index(original_refit, refit_metric, cv_results_)
                if not callable(original_refit):
                    # With a non-custom callable, we can select the best score
                    # based on the best index
                    estimator.best_score_ = cv_results_[f"mean_test_{refit_metric}"][estimator.best_index_]
                estimator.best_params_ = cv_results_["params"][best_param_index]

            if original_refit:
                estimator.best_estimator_ = clone(estimator.estimator).set_params(
                    **clone(estimator.best_params_, safe=False)
                )

                # Let the sproc use all cores to refit.
                estimator.n_jobs = -1 if not estimator.n_jobs else estimator.n_jobs

                # process the input as args
                argspec = inspect.getfullargspec(estimator.fit)
                args = {"X": X}
                if label_cols:
                    label_arg_name = "Y" if "Y" in argspec.args else "y"
                    args[label_arg_name] = y
                if sample_weight_col is not None and "sample_weight" in argspec.args:
                    args["sample_weight"] = df[sample_weight_col].squeeze()
                estimator.refit = original_refit
                refit_start_time = time.time()
                estimator.best_estimator_.fit(**args)
                refit_end_time = time.time()
                estimator.refit_time_ = refit_end_time - refit_start_time

                if hasattr(estimator.best_estimator_, "feature_names_in_"):
                    estimator.feature_names_in_ = estimator.best_estimator_.feature_names_in_

            local_result_file_name = get_temp_file_path()

            with open(local_result_file_name, mode="w+b") as local_result_file_obj:
                cp.dump(estimator, local_result_file_obj)

            session.file.put(
                local_result_file_name,
                temp_stage_name,
                auto_compress=False,
                overwrite=True,
            )

            # Note: you can add something like  + "|" + str(df) to the return string
            # to pass debug information to the caller.
            return str(os.path.basename(local_result_file_name))

        sproc_export_file_name = _distributed_search(
            session,
            imports,
            stage_estimator_file_name,
            input_cols,
            label_cols,
        )

        local_estimator_path = get_temp_file_path()
        session.file.get(
            posixpath.join(temp_stage_name, sproc_export_file_name),
            local_estimator_path,
        )

        with open(os.path.join(local_estimator_path, sproc_export_file_name), mode="r+b") as result_file_obj:
            fit_estimator = cp.load(result_file_obj)

        cleanup_temp_files([local_estimator_path])

        return fit_estimator



import inspect
from typing import List, Optional, Union

import cloudpickle
import numpy as np
import numpy.typing as npt
import sklearn
from packaging import version
from sklearn import metrics

import snowflake.snowpark._internal.utils as snowpark_utils
from snowflake import snowpark
from snowflake.ml._internal import telemetry
from snowflake.ml._internal.utils import result
from snowflake.ml.modeling.metrics import metrics_utils
from snowflake.snowpark import functions as F

_PROJECT = "ModelDevelopment"
_SUBPROJECT = "Metrics"
_MULTIOUTPUT_UNIFORM_AVG = "uniform_average"
_MULTIOUTPUT_RAW_VALUES = "raw_values"


@telemetry.send_api_usage_telemetry(project=_PROJECT, subproject=_SUBPROJECT)
def d2_absolute_error_score(
    *,
    df: snowpark.DataFrame,
    y_true_col_names: Union[str, List[str]],
    y_pred_col_names: Union[str, List[str]],
    sample_weight_col_name: Optional[str] = None,
    multioutput: Union[str, npt.ArrayLike] = "uniform_average",
) -> Union[float, npt.NDArray[np.float_]]:
    """
    :math:`D^2` regression score function, \
    fraction of absolute error explained.

    Best possible score is 1.0 and it can be negative (because the model can be
    arbitrarily worse). A model that always uses the empirical median of `y_true`
    as constant prediction, disregarding the input features,
    gets a :math:`D^2` score of 0.0.

    Args:
        df: Input dataframe.
        y_true_col_names: Column name(s) representing actual values.
        y_pred_col_names: Column name(s) representing predicted values.
        sample_weight_col_name: Column name representing sample weights.
        multioutput: {'raw_values', 'uniform_average'}  or array-like of shape \
            (n_outputs,), default='uniform_average'
            Defines aggregating of multiple output values.
            Array-like value defines weights used to average errors.
            'raw_values':
                Returns a full set of errors in case of multioutput input.
            'uniform_average':
                Errors of all outputs are averaged with uniform weight.

    Returns:
        score: float or ndarray of floats
            The :math:`D^2` score with an absolute error deviance
            or ndarray of scores if 'multioutput' is 'raw_values'.
    """
    metrics_utils.check_label_columns(y_true_col_names, y_pred_col_names)

    session = df._session
    assert session is not None
    sproc_name = snowpark_utils.random_name_for_temp_object(snowpark_utils.TempObjectType.PROCEDURE)
    sklearn_release = version.parse(sklearn.__version__).release
    statement_params = telemetry.get_statement_params(_PROJECT, _SUBPROJECT)

    cols = metrics_utils.flatten_cols([y_true_col_names, y_pred_col_names, sample_weight_col_name])
    queries = df[cols].queries["queries"]

    pickled_snowflake_result = cloudpickle.dumps(result)

    @F.sproc(  # type: ignore[misc]
        is_permanent=False,
        session=session,
        name=sproc_name,
        replace=True,
        packages=[
            "cloudpickle",
            f"scikit-learn=={sklearn_release[0]}.{sklearn_release[1]}.*",
            "snowflake-snowpark-python",
        ],
        statement_params=statement_params,
        anonymous=True,
    )
    def d2_absolute_error_score_anon_sproc(session: snowpark.Session) -> bytes:
        for query in queries[:-1]:
            _ = session.sql(query).collect(statement_params=statement_params)
        sp_df = session.sql(queries[-1])
        df = sp_df.to_pandas(statement_params=statement_params)
        df.columns = sp_df.columns

        y_true = df[y_true_col_names]
        y_pred = df[y_pred_col_names]
        sample_weight = df[sample_weight_col_name] if sample_weight_col_name else None

        score = metrics.d2_absolute_error_score(
            y_true,
            y_pred,
            sample_weight=sample_weight,
            multioutput=multioutput,
        )
        result_module = cloudpickle.loads(pickled_snowflake_result)
        return result_module.serialize(session, score)  # type: ignore[no-any-return]

    result_object = result.deserialize(session, d2_absolute_error_score_anon_sproc(session))
    score: Union[float, npt.NDArray[np.float_]] = result_object
    return score


@telemetry.send_api_usage_telemetry(project=_PROJECT, subproject=_SUBPROJECT)
def d2_pinball_score(
    *,
    df: snowpark.DataFrame,
    y_true_col_names: Union[str, List[str]],
    y_pred_col_names: Union[str, List[str]],
    sample_weight_col_name: Optional[str] = None,
    alpha: float = 0.5,
    multioutput: Union[str, npt.ArrayLike] = "uniform_average",
) -> Union[float, npt.NDArray[np.float_]]:
    """
    :math:`D^2` regression score function, fraction of pinball loss explained.

    Best possible score is 1.0 and it can be negative (because the model can be
    arbitrarily worse). A model that always uses the empirical alpha-quantile of
    `y_true` as constant prediction, disregarding the input features,
    gets a :math:`D^2` score of 0.0.

    Args:
        df: Input dataframe.
        y_true_col_names: Column name(s) representing actual values.
        y_pred_col_names: Column name(s) representing predicted values.
        sample_weight_col_name: Column name representing sample weights.
        alpha: Slope of the pinball deviance. It determines the quantile level
            alpha for which the pinball deviance and also D2 are optimal.
            The default `alpha=0.5` is equivalent to `d2_absolute_error_score`.
        multioutput: {'raw_values', 'uniform_average'}  or array-like of shape \
            (n_outputs,), default='uniform_average'
            Defines aggregating of multiple output values.
            Array-like value defines weights used to average errors.
            'raw_values':
                Returns a full set of errors in case of multioutput input.
            'uniform_average':
                Scores of all outputs are averaged with uniform weight.

    Returns:
        score: float or ndarray of floats
            The :math:`D^2` score with a pinball deviance
            or ndarray of scores if `multioutput='raw_values'`.
    """
    metrics_utils.check_label_columns(y_true_col_names, y_pred_col_names)

    session = df._session
    assert session is not None
    sproc_name = snowpark_utils.random_name_for_temp_object(snowpark_utils.TempObjectType.PROCEDURE)
    sklearn_release = version.parse(sklearn.__version__).release
    statement_params = telemetry.get_statement_params(_PROJECT, _SUBPROJECT)

    cols = metrics_utils.flatten_cols([y_true_col_names, y_pred_col_names, sample_weight_col_name])
    queries = df[cols].queries["queries"]

    pickled_result_module = cloudpickle.dumps(result)

    @F.sproc(  # type: ignore[misc]
        is_permanent=False,
        session=session,
        name=sproc_name,
        replace=True,
        packages=[
            "cloudpickle",
            f"scikit-learn=={sklearn_release[0]}.{sklearn_release[1]}.*",
            "snowflake-snowpark-python",
        ],
        statement_params=statement_params,
        anonymous=True,
    )
    def d2_pinball_score_anon_sproc(session: snowpark.Session) -> bytes:
        for query in queries[:-1]:
            _ = session.sql(query).collect(statement_params=statement_params)
        sp_df = session.sql(queries[-1])
        df = sp_df.to_pandas(statement_params=statement_params)
        df.columns = sp_df.columns

        y_true = df[y_true_col_names]
        y_pred = df[y_pred_col_names]
        sample_weight = df[sample_weight_col_name] if sample_weight_col_name else None

        score = metrics.d2_pinball_score(
            y_true,
            y_pred,
            sample_weight=sample_weight,
            alpha=alpha,
            multioutput=multioutput,
        )
        result_module = cloudpickle.loads(pickled_result_module)
        return result_module.serialize(session, score)  # type: ignore[no-any-return]

    result_object = result.deserialize(session, d2_pinball_score_anon_sproc(session))

    score: Union[float, npt.NDArray[np.float_]] = result_object
    return score


@telemetry.send_api_usage_telemetry(project=_PROJECT, subproject=_SUBPROJECT)
def explained_variance_score(
    *,
    df: snowpark.DataFrame,
    y_true_col_names: Union[str, List[str]],
    y_pred_col_names: Union[str, List[str]],
    sample_weight_col_name: Optional[str] = None,
    multioutput: Union[str, npt.ArrayLike] = "uniform_average",
    force_finite: bool = True,
) -> Union[float, npt.NDArray[np.float_]]:
    """
    Explained variance regression score function.

    Best possible score is 1.0, lower values are worse.

    In the particular case when ``y_true`` is constant, the explained variance
    score is not finite: it is either ``NaN`` (perfect predictions) or
    ``-Inf`` (imperfect predictions). To prevent such non-finite numbers to
    pollute higher-level experiments such as a grid search cross-validation,
    by default these cases are replaced with 1.0 (perfect predictions) or 0.0
    (imperfect predictions) respectively. If ``force_finite``
    is set to ``False``, this score falls back on the original :math:`R^2`
    definition.

    Note:
       The Explained Variance score is similar to the
       :func:`R^2 score <r2_score>`, with the notable difference that it
       does not account for systematic offsets in the prediction. Most often
       the :func:`R^2 score <r2_score>` should be preferred.

    Args:
        df: Input dataframe.
        y_true_col_names: Column name(s) representing actual values.
        y_pred_col_names: Column name(s) representing predicted values.
        sample_weight_col_name: Column name representing sample weights.
        multioutput: {'raw_values', 'uniform_average', 'variance_weighted'} or \
            array-like of shape (n_outputs,), default='uniform_average'
            Defines aggregating of multiple output values.
            Array-like value defines weights used to average errors.
            'raw_values':
                Returns a full set of scores in case of multioutput input.
            'uniform_average':
                Scores of all outputs are averaged with uniform weight.
            'variance_weighted':
                Scores of all outputs are averaged, weighted by the variances
                of each individual output.
        force_finite: Flag indicating if ``NaN`` and ``-Inf`` scores resulting
            from constant data should be replaced with real numbers (``1.0`` if
            prediction is perfect, ``0.0`` otherwise). Default is ``True``, a
            convenient setting for hyperparameters' search procedures (e.g. grid
            search cross-validation).

    Returns:
        score: float or ndarray of floats
            The explained variance or ndarray if 'multioutput' is 'raw_values'.
    """
    metrics_utils.check_label_columns(y_true_col_names, y_pred_col_names)

    session = df._session
    assert session is not None
    sproc_name = snowpark_utils.random_name_for_temp_object(snowpark_utils.TempObjectType.PROCEDURE)
    sklearn_release = version.parse(sklearn.__version__).release
    statement_params = telemetry.get_statement_params(_PROJECT, _SUBPROJECT)

    cols = metrics_utils.flatten_cols([y_true_col_names, y_pred_col_names, sample_weight_col_name])
    queries = df[cols].queries["queries"]

    pickled_result_module = cloudpickle.dumps(result)

    @F.sproc(  # type: ignore[misc]
        is_permanent=False,
        session=session,
        name=sproc_name,
        replace=True,
        packages=[
            "cloudpickle",
            f"scikit-learn=={sklearn_release[0]}.{sklearn_release[1]}.*",
            "snowflake-snowpark-python",
        ],
        statement_params=statement_params,
        anonymous=True,
    )
    def explained_variance_score_anon_sproc(session: snowpark.Session) -> bytes:
        for query in queries[:-1]:
            _ = session.sql(query).collect(statement_params=statement_params)
        sp_df = session.sql(queries[-1])
        df = sp_df.to_pandas(statement_params=statement_params)
        df.columns = sp_df.columns

        y_true = df[y_true_col_names]
        y_pred = df[y_pred_col_names]
        sample_weight = df[sample_weight_col_name] if sample_weight_col_name else None

        score = metrics.explained_variance_score(
            y_true,
            y_pred,
            sample_weight=sample_weight,
            multioutput=multioutput,
            force_finite=force_finite,
        )
        result_module = cloudpickle.loads(pickled_result_module)
        return result_module.serialize(session, score)  # type: ignore[no-any-return]

    result_object = result.deserialize(session, explained_variance_score_anon_sproc(session))
    score: Union[float, npt.NDArray[np.float_]] = result_object
    return score


@telemetry.send_api_usage_telemetry(project=_PROJECT, subproject=_SUBPROJECT)
def mean_absolute_error(
    *,
    df: snowpark.DataFrame,
    y_true_col_names: Union[str, List[str]],
    y_pred_col_names: Union[str, List[str]],
    sample_weight_col_name: Optional[str] = None,
    multioutput: Union[str, npt.ArrayLike] = "uniform_average",
) -> Union[float, npt.NDArray[np.float_]]:
    """
    Mean absolute error regression loss.

    Args:
        df: Input dataframe.
        y_true_col_names: Column name(s) representing actual values.
        y_pred_col_names: Column name(s) representing predicted values.
        sample_weight_col_name: Column name representing sample weights.
        multioutput: {'raw_values', 'uniform_average'}  or array-like of shape \
            (n_outputs,), default='uniform_average'
            Defines aggregating of multiple output values.
            Array-like value defines weights used to average errors.
            'raw_values':
                Returns a full set of errors in case of multioutput input.
            'uniform_average':
                Errors of all outputs are averaged with uniform weight.

    Returns:
        loss: float or ndarray of floats
            If multioutput is 'raw_values', then mean absolute error is returned
            for each output separately.
            If multioutput is 'uniform_average' or an ndarray of weights, then the
            weighted average of all output errors is returned.

            MAE output is non-negative floating point. The best value is 0.0.
    """
    metrics_utils.check_label_columns(y_true_col_names, y_pred_col_names)
    session = df._session
    assert session is not None
    statement_params = telemetry.get_statement_params(_PROJECT, _SUBPROJECT)
    y_true = y_true_col_names if isinstance(y_true_col_names, list) else [y_true_col_names]
    y_pred = y_pred_col_names if isinstance(y_pred_col_names, list) else [y_pred_col_names]
    num_outputs = len(y_true)
    _validate_multioutput(multioutput, num_outputs)

    for i in range(num_outputs):
        df = df.with_column(
            "diff" + str(i),
            F.abs(df[y_pred[i]] - df[y_true[i]]),
        )
    output_errors = np.array(
        [
            metrics_utils.weighted_sum(
                df=df,
                sample_score_column=df["diff" + str(i)],
                sample_weight_column=df[sample_weight_col_name] if sample_weight_col_name else None,
                normalize=True,
                statement_params=statement_params,
            )
            for i in range(num_outputs)
        ]
    )

    if multioutput == _MULTIOUTPUT_RAW_VALUES:
        return output_errors

    weights = None if multioutput == _MULTIOUTPUT_UNIFORM_AVG else multioutput
    return float(np.average(output_errors, weights=weights))


@telemetry.send_api_usage_telemetry(project=_PROJECT, subproject=_SUBPROJECT)
def mean_absolute_percentage_error(
    *,
    df: snowpark.DataFrame,
    y_true_col_names: Union[str, List[str]],
    y_pred_col_names: Union[str, List[str]],
    sample_weight_col_name: Optional[str] = None,
    multioutput: Union[str, npt.ArrayLike] = "uniform_average",
) -> Union[float, npt.NDArray[np.float_]]:
    """
    Mean absolute percentage error (MAPE) regression loss.

    Note here that the output is not a percentage in the range [0, 100]
    and a value of 100 does not mean 100% but 1e2. Furthermore, the output
    can be arbitrarily high when `y_true` is small (which is specific to the
    metric) or when `abs(y_true - y_pred)` is large (which is common for most
    regression metrics).

    Args:
        df: Input dataframe.
        y_true_col_names: Column name(s) representing actual values.
        y_pred_col_names: Column name(s) representing predicted values.
        sample_weight_col_name: Column name representing sample weights.
        multioutput: {'raw_values', 'uniform_average'}  or array-like of shape \
            (n_outputs,), default='uniform_average'
            Defines aggregating of multiple output values.
            Array-like value defines weights used to average errors.
            'raw_values':
                Returns a full set of errors in case of multioutput input.
            'uniform_average':
                Errors of all outputs are averaged with uniform weight.

    Returns:
        loss: float or ndarray of floats
            If multioutput is 'raw_values', then mean absolute percentage error
            is returned for each output separately.
            If multioutput is 'uniform_average' or an ndarray of weights, then the
            weighted average of all output errors is returned.

            MAPE output is non-negative floating point. The best value is 0.0.
            But note that bad predictions can lead to arbitrarily large
            MAPE values, especially if some `y_true` values are very close to zero.
            Note that we return a large value instead of `inf` when `y_true` is zero.
    """
    metrics_utils.check_label_columns(y_true_col_names, y_pred_col_names)
    session = df._session
    assert session is not None
    statement_params = telemetry.get_statement_params(_PROJECT, _SUBPROJECT)
    y_true = y_true_col_names if isinstance(y_true_col_names, list) else [y_true_col_names]
    y_pred = y_pred_col_names if isinstance(y_pred_col_names, list) else [y_pred_col_names]
    num_outputs = len(y_true)
    _validate_multioutput(multioutput, num_outputs)

    epsilon = float(np.finfo(np.float64).eps)
    for i in range(num_outputs):
        df = df.with_column(
            "MAPE" + str(i),
            F.abs(df[y_pred[i]] - df[y_true[i]]) / F.iff(F.abs(df[y_true[i]]) > epsilon, F.abs(df[y_true[i]]), epsilon),
        )
    output_errors = np.array(
        [
            metrics_utils.weighted_sum(
                df=df,
                sample_score_column=df["MAPE" + str(i)],
                sample_weight_column=df[sample_weight_col_name] if sample_weight_col_name else None,
                normalize=True,
                statement_params=statement_params,
            )
            for i in range(num_outputs)
        ]
    )

    if multioutput == _MULTIOUTPUT_RAW_VALUES:
        return output_errors

    weights = None if multioutput == _MULTIOUTPUT_UNIFORM_AVG else multioutput
    return float(np.average(output_errors, weights=weights))


@telemetry.send_api_usage_telemetry(project=_PROJECT, subproject=_SUBPROJECT)
def mean_squared_error(
    *,
    df: snowpark.DataFrame,
    y_true_col_names: Union[str, List[str]],
    y_pred_col_names: Union[str, List[str]],
    sample_weight_col_name: Optional[str] = None,
    multioutput: Union[str, npt.ArrayLike] = "uniform_average",
    squared: bool = True,
) -> Union[float, npt.NDArray[np.float_]]:
    """
    Mean squared error regression loss.

    Args:
        df: Input dataframe.
        y_true_col_names: Column name(s) representing actual values.
        y_pred_col_names: Column name(s) representing predicted values.
        sample_weight_col_name: Column name representing sample weights.
        multioutput: {'raw_values', 'uniform_average'}  or array-like of shape \
            (n_outputs,), default='uniform_average'
            Defines aggregating of multiple output values.
            Array-like value defines weights used to average errors.
            'raw_values':
                Returns a full set of errors in case of multioutput input.
            'uniform_average':
                Errors of all outputs are averaged with uniform weight.
        squared: If True returns MSE value, if False returns RMSE value.

    Returns:
        loss: float or ndarray of floats
            A non-negative floating point value (the best value is 0.0), or an
            array of floating point values, one for each individual target.
    """
    metrics_utils.check_label_columns(y_true_col_names, y_pred_col_names)
    session = df._session
    assert session is not None
    statement_params = telemetry.get_statement_params(_PROJECT, _SUBPROJECT)
    y_true = y_true_col_names if isinstance(y_true_col_names, list) else [y_true_col_names]
    y_pred = y_pred_col_names if isinstance(y_pred_col_names, list) else [y_pred_col_names]
    num_outputs = len(y_true)
    _validate_multioutput(multioutput, num_outputs)

    for i in range(num_outputs):
        df = df.with_column(
            "diff" + str(i),
            F.pow(df[y_true[i]] - df[y_pred[i]], 2),
        )

    output_errors = np.array(
        [
            metrics_utils.weighted_sum(
                df=df,
                sample_score_column=df["diff" + str(i)],
                sample_weight_column=df[sample_weight_col_name] if sample_weight_col_name else None,
                normalize=True,
                statement_params=statement_params,
            )
            for i in range(num_outputs)
        ]
    )
    if not squared:
        output_errors = np.sqrt(output_errors)

    if multioutput == _MULTIOUTPUT_RAW_VALUES:
        return output_errors

    weights = None if multioutput == _MULTIOUTPUT_UNIFORM_AVG else multioutput
    return float(np.average(output_errors, weights=weights))


@telemetry.send_api_usage_telemetry(project=_PROJECT, subproject=_SUBPROJECT)
def r2_score(*, df: snowpark.DataFrame, y_true_col_name: str, y_pred_col_name: str) -> float:
    """:math:`R^2` (coefficient of determination) regression score function.
    Returns R squared metric on 2 columns in the dataframe.

    Best possible score is 1.0 and it can be negative (because the
    model can be arbitrarily worse). In the general case when the true y is
    non-constant, a constant model that always predicts the average y
    disregarding the input features would get a :math:`R^2` score of 0.0.

    TODO(pdorairaj): Implement other params from sklearn - sample_weight, multi_output, force_finite.

    Args:
        df: Input dataframe.
        y_true_col_name: Column name representing actual values.
        y_pred_col_name: Column name representing predicted values.

    Returns:
        R squared metric.
    """

    df_avg = df.select(F.avg(y_true_col_name).as_("avg_y_true"))
    df_r_square = df.join(df_avg).select(
        F.lit(1)
        - F.sum((df[y_true_col_name] - df[y_pred_col_name]) ** 2)
        / F.sum((df[y_true_col_name] - df_avg["avg_y_true"]) ** 2)
    )

    statement_params = telemetry.get_function_usage_statement_params(
        project=_PROJECT,
        subproject=_SUBPROJECT,
        function_name=telemetry.get_statement_params_full_func_name(inspect.currentframe(), None),
    )
    return float(df_r_square.collect(statement_params=statement_params)[0][0])


def _validate_multioutput(multioutput: Union[str, npt.ArrayLike], num_outputs: int) -> None:
    """Validates multioutput parameter for MAPE calculation.

    Args:
        multioutput: Parameter specifying how to deal with multiple outputs.
        num_outputs: Integer representing number of outputs.

    Raises:
        ValueError: multioutput parameter is invalid.

    """
    allowed_str_vals = ("raw_values", "uniform_average")
    if isinstance(multioutput, str):
        if multioutput not in allowed_str_vals:
            raise ValueError(
                f"Allowed 'multioutput' string values are {allowed_str_vals}. You provided multioutput={multioutput}"
            )
    else:
        multioutput_np = np.array(multioutput)
        if num_outputs == 1:
            raise ValueError("Custom weights are useful only in multi-output cases.")
        elif num_outputs != len(multioutput_np):
            raise ValueError(
                f"There must be equally many custom weights ({len(multioutput_np)}) as outputs ({num_outputs})."
            )



from typing import Any, Dict, List, Optional, Tuple

from snowflake import snowpark
from snowflake.ml._internal.utils import identifier, query_result_checker, table_manager
from snowflake.ml.registry import _initial_schema, _schema

_SCHEMA_VERSION_TABLE_NAME: str = "_SYSTEM_REGISTRY_SCHEMA_VERSION"

_SCHEMA_VERSION_TABLE_SCHEMA: List[Tuple[str, str]] = [
    ("VERSION", "NUMBER"),
    ("CREATION_TIME", "TIMESTAMP_TZ"),
    ("INFO", "OBJECT"),
]


class SchemaVersionManager:
    """Registry Schema Version Manager to deal with schema evolution."""

    def __init__(self, session: snowpark.Session, database: str, schema: str) -> None:
        """SchemaVersionManager constructor.

        Args:
            session: Snowpark session
            database: Database in which registry is being managed.
            schema: Schema in which registry is being managed.
        """
        self._session = session
        self._database = database
        self._schema = schema

    def get_deployed_version(self, statement_params: Optional[Dict[str, Any]] = None) -> int:
        """Get current version of deployed schema.

        Args:
            statement_params: Statement parameters for telemetry tracking.

        Returns:
            Version of deployed schema.
        """
        if not table_manager.validate_table_exist(
            self._session, _SCHEMA_VERSION_TABLE_NAME, self._get_qualified_schema()
        ):
            return _initial_schema._INITIAL_VERSION

        result = (
            query_result_checker.SqlResultValidator(
                session=self._session,
                query=f"""SELECT MAX(VERSION) AS MAX_VERSION
                        FROM {self._get_qualified_schema()}.{_SCHEMA_VERSION_TABLE_NAME}
                        """,
                statement_params=statement_params,
            )
            .has_dimensions(expected_rows=1, expected_cols=1)
            .has_column("MAX_VERSION")
            .validate()
        )
        cur_version = result[0]["MAX_VERSION"]
        return int(cur_version)

    def validate_schema_version(self, statement_params: Optional[Dict[str, Any]] = None) -> None:
        """Checks if currently deployed schema is up to date.

        Args:
            statement_params: Statement parameters for telemetry tracking.

        Raises:
            RuntimeError: if deployed schema different from registry schema.
        """
        deployed_version = self.get_deployed_version(statement_params)
        if deployed_version > _schema._CURRENT_SCHEMA_VERSION:
            raise RuntimeError(
                f"Deployed registry schema version ({deployed_version}) is ahead of current "
                f"package ({_schema._CURRENT_SCHEMA_VERSION}). Please update the package."
            )
        elif deployed_version < _schema._CURRENT_SCHEMA_VERSION:
            raise RuntimeError(
                f"Registry schema version ({_schema._CURRENT_SCHEMA_VERSION}) is ahead of deployed "
                f"schema ({deployed_version}). Please call create_model_registry() to upgrade."
            )

    def try_upgrade(self, statement_params: Optional[Dict[str, Any]] = None) -> None:
        """Upgrade deployed schema to current.

        Args:
            statement_params: Statement parameters for telemetry tracking.

        Raises:
            RuntimeError: Deployed schema is newer than package.
        """
        deployed_version = self.get_deployed_version(statement_params)
        if deployed_version > _schema._CURRENT_SCHEMA_VERSION:
            raise RuntimeError(
                f"Deployed registry schema version ({deployed_version}) is ahead of current "
                f"package ({_schema._CURRENT_SCHEMA_VERSION}). Please update the package."
            )

        any_upgraded = False
        for cur_version in range(deployed_version, _schema._CURRENT_SCHEMA_VERSION):
            assert cur_version + 1 in _schema._SCHEMA_UPGRADE_PLANS, "version number not exist."
            plan = _schema._SCHEMA_UPGRADE_PLANS[cur_version + 1](
                self._session, self._database, self._schema, statement_params
            )
            plan.upgrade()
            any_upgraded = True

        self._validate_schema()

        if any_upgraded:
            self._create_or_update_version_table(statement_params)

    def _create_or_update_version_table(self, statement_params: Optional[Dict[str, Any]] = None) -> None:
        if not table_manager.validate_table_exist(
            self._session, _SCHEMA_VERSION_TABLE_NAME, self._get_qualified_schema()
        ):
            table_manager.create_single_table(
                session=self._session,
                database_name=self._database,
                schema_name=self._schema,
                table_name=_SCHEMA_VERSION_TABLE_NAME,
                table_schema=_SCHEMA_VERSION_TABLE_SCHEMA,
                statement_params=statement_params,
            )
        query_result_checker.SqlResultValidator(
            session=self._session,
            query=f"""INSERT INTO {self._get_qualified_schema_version_table()} (VERSION, CREATION_TIME)
                VALUES ({_schema._CURRENT_SCHEMA_VERSION}, CURRENT_TIMESTAMP())
            """,
            statement_params=statement_params,
        ).insertion_success(expected_num_rows=1).validate()

    def _validate_schema(self) -> None:
        for table_name in _initial_schema._INITIAL_TABLE_SCHEMAS:
            if table_name not in _schema._CURRENT_TABLE_SCHEMAS:
                # This table must be deleted by transformations.
                if table_manager.validate_table_exist(
                    self._session,
                    table_name,
                    table_manager.get_fully_qualified_schema_name(self._database, self._schema),
                ):
                    raise RuntimeError(
                        f"Schema transformation error. A table '{table_name}' found, which should not exist."
                    )

        exclude_cols = ["PRIMARY KEY"]
        for table_name, expected_schema in _schema._CURRENT_TABLE_SCHEMAS.items():
            deployed_schema_dict = table_manager.get_table_schema(
                self._session, table_name, self._get_qualified_schema()
            )

            # TODO check type as well.
            for col_name, _ in expected_schema:
                if col_name not in deployed_schema_dict and col_name not in exclude_cols:
                    raise RuntimeError(f"Schema table: {table_name} doesn't have required column:'{col_name}'.")

    def _get_qualified_schema(self) -> str:
        return table_manager.get_fully_qualified_schema_name(self._database, self._schema)

    def _get_qualified_schema_version_table(self) -> str:
        return table_manager.get_fully_qualified_table_name(
            self._database,
            self._schema,
            identifier.get_inferred_name(_SCHEMA_VERSION_TABLE_NAME),
        )



from typing import List

from snowflake.ml._internal.utils import identifier


class SqlIdentifier(str):
    """Represents an identifier in SQL. An identifier has 3 states:
        1. User input: this is the raw input string to initializer.
        2. identifier(): this is the state that ready input to SQL.
        3. resolved(): this is the state how the identifier stored in database.

    For example:
        1. user input                           ->    2. identifier()     -> 3. resolved()
        SqlIdentifier('abc', case_sensitive=False)          ABC                 ABC
        SqlIdentifier('"abc"', case_sensitive=False)        "abc"               abc
        SqlIdentifier('abc', case_sensitive=True)           "abc"               abc
    """

    def __new__(cls, name: str, *, case_sensitive: bool = False) -> "SqlIdentifier":
        """Create new instance of sql identifier.
            Refer to here for more details: https://docs.snowflake.com/en/sql-reference/identifiers-syntax

        Args:
            name: A string name.
            case_sensitive: If False, then the input string is considered case insensitive and will follow SQL
                identifier parsing rule; if True, then the input string is considered case sensitive, so quotes are
                automatically added if necessary to make sure the original input's cases are preserved.
                Default to False.

        Raises:
            ValueError: input name is not a valid identifier.

        Returns:
            Returns new instance created.
        """
        # TODO (wezhou) add stronger validation to recognize a valid snowflake identifier.
        if not name:
            raise ValueError(f"name:`{name}` is not a valid identifier.")
        if case_sensitive:
            return super().__new__(cls, identifier.get_inferred_name(name))
        else:
            return super().__new__(cls, identifier.resolve_identifier(name))

    def __init__(self, name: str, case_sensitive: bool = False) -> None:
        """Initialize sql identifier.

        Args:
            name: A string name.
            case_sensitive: If False, then the input string is considered case insensitive and will follow SQL
                identifier parsing rule; if True, then the input string is considered case sensitive, so quotes are
                automatically added if necessary to make sure the original input's cases are preserved.
                Default to False.
        """
        super().__init__()

    def identifier(self) -> str:
        """Get the identifier value. This is how the string looks like input to SQL.

        Returns:
            An identifier string.
        """
        return str(self)

    def resolved(self) -> str:
        """Get a resolved string after applying identifier requirement rules. This is how the identifier stored
            in database.

        Returns:
            A resolved string.
        """
        return identifier.get_unescaped_names(str(self))

    def __eq__(self, other: object) -> bool:
        if isinstance(other, SqlIdentifier):
            return self.resolved() == other.resolved()
        if isinstance(other, str):
            return str(self) == other
        return False

    def __hash__(self) -> int:
        return super().__hash__()


def to_sql_identifiers(list_of_str: List[str], *, case_sensitive: bool = False) -> List[SqlIdentifier]:
    return [SqlIdentifier(val, case_sensitive=case_sensitive) for val in list_of_str]



import copy
import logging
import os
import posixpath
import string
import tempfile
import time
from contextlib import contextmanager
from typing import Any, Dict, Generator, Optional, cast

import importlib_resources
import yaml
from typing_extensions import Unpack

from snowflake.ml._internal import file_utils
from snowflake.ml._internal.exceptions import (
    error_codes,
    exceptions as snowml_exceptions,
)
from snowflake.ml._internal.utils import identifier, query_result_checker
from snowflake.ml.model import type_hints
from snowflake.ml.model._deploy_client import snowservice
from snowflake.ml.model._deploy_client.image_builds import (
    base_image_builder,
    client_image_builder,
    docker_context,
    server_image_builder,
)
from snowflake.ml.model._deploy_client.snowservice import deploy_options, instance_types
from snowflake.ml.model._deploy_client.utils import (
    constants,
    image_registry_client,
    snowservice_client,
)
from snowflake.ml.model._packager.model_meta import model_meta, model_meta_schema
from snowflake.snowpark import Session

logger = logging.getLogger(__name__)


@contextmanager
def _debug_aware_tmp_directory(debug_dir: Optional[str] = None) -> Generator[str, None, None]:
    """Debug-aware directory provider.

    Args:
        debug_dir: A folder for deploymement context.

    Yields:
        A directory path to write deployment artifacts
    """
    create_temp = False
    if debug_dir:
        directory_path = debug_dir
    else:
        temp_dir_context = tempfile.TemporaryDirectory()
        directory_path = temp_dir_context.name
        create_temp = True
    try:
        yield directory_path
    finally:
        if create_temp:
            temp_dir_context.cleanup()


def _deploy(
    session: Session,
    *,
    model_id: str,
    model_meta: model_meta.ModelMetadata,
    service_func_name: str,
    model_zip_stage_path: str,
    deployment_stage_path: str,
    target_method: str,
    **kwargs: Unpack[type_hints.SnowparkContainerServiceDeployOptions],
) -> type_hints.SnowparkContainerServiceDeployDetails:
    """Entrypoint for model deployment to SnowService. This function will trigger a docker image build followed by
    workflow deployment to SnowService.

    Args:
        session: Snowpark session
        model_id: Unique hex string of length 32, provided by model registry.
        model_meta: Model Metadata.
        service_func_name: The service function name in SnowService associated with the created service.
        model_zip_stage_path: Path to model zip file in stage. Note that this path has a "@" prefix.
        deployment_stage_path: Path to stage containing deployment artifacts.
        target_method: The name of the target method to be deployed.
        **kwargs: various SnowService deployment options.

    Returns:
        Deployment details for SPCS.

    Raises:
        SnowflakeMLException: Raised when model_id is empty.
        SnowflakeMLException: Raised when service_func_name is empty.
        SnowflakeMLException: Raised when model_stage_file_path is empty.
    """
    snowpark_logger = logging.getLogger("snowflake.snowpark")
    snowflake_connector_logger = logging.getLogger("snowflake.connector")
    snowpark_log_level = snowpark_logger.level
    snowflake_connector_log_level = snowflake_connector_logger.level

    query_result = (
        query_result_checker.SqlResultValidator(
            session,
            query="SHOW PARAMETERS LIKE 'PYTHON_CONNECTOR_QUERY_RESULT_FORMAT' IN SESSION",
        )
        .has_dimensions(expected_rows=1)
        .validate()
    )
    prev_format = query_result[0].value

    try:
        # Setting appropriate log level to prevent console from being polluted by vast amount of snowpark and snowflake
        # connector logging.
        snowpark_logger.setLevel(logging.WARNING)
        snowflake_connector_logger.setLevel(logging.WARNING)

        # Query format change is needed to ensure session token obtained from the session object is valid.
        session.sql("ALTER SESSION SET PYTHON_CONNECTOR_QUERY_RESULT_FORMAT = 'json'").collect()
        if not model_id:
            raise snowml_exceptions.SnowflakeMLException(
                error_code=error_codes.INVALID_ARGUMENT,
                original_exception=ValueError(
                    'Must provide a non-empty string for "model_id" when deploying to Snowpark Container Services'
                ),
            )
        if not service_func_name:
            raise snowml_exceptions.SnowflakeMLException(
                error_code=error_codes.INVALID_ARGUMENT,
                original_exception=ValueError(
                    'Must provide a non-empty string for "service_func_name"'
                    " when deploying to Snowpark Container Services"
                ),
            )
        if not model_zip_stage_path:
            raise snowml_exceptions.SnowflakeMLException(
                error_code=error_codes.INVALID_ARGUMENT,
                original_exception=ValueError(
                    'Must provide a non-empty string for "model_stage_file_path"'
                    " when deploying to Snowpark Container Services"
                ),
            )
        if not deployment_stage_path:
            raise snowml_exceptions.SnowflakeMLException(
                error_code=error_codes.INVALID_ARGUMENT,
                original_exception=ValueError(
                    'Must provide a non-empty string for "deployment_stage_path"'
                    " when deploying to Snowpark Container Services"
                ),
            )

        # Remove full qualified name to avoid double quotes corrupting the service spec
        model_zip_stage_path = model_zip_stage_path.replace('"', "")
        deployment_stage_path = deployment_stage_path.replace('"', "")

        assert model_zip_stage_path.startswith("@"), f"stage path should start with @, actual: {model_zip_stage_path}"
        assert deployment_stage_path.startswith("@"), f"stage path should start with @, actual: {deployment_stage_path}"
        options = deploy_options.SnowServiceDeployOptions.from_dict(cast(Dict[str, Any], kwargs))

        model_meta_deploy = copy.deepcopy(model_meta)
        # Set conda-forge as backup channel for SPCS deployment
        if "conda-forge" not in model_meta_deploy.env._conda_dependencies:
            model_meta_deploy.env._conda_dependencies["conda-forge"] = []
        if options.use_gpu:
            # Make mypy happy
            assert options.num_gpus is not None
            if model_meta_deploy.env.cuda_version is None:
                raise snowml_exceptions.SnowflakeMLException(
                    error_code=error_codes.INVALID_ARGUMENT,
                    original_exception=ValueError(
                        "You are requesting GPUs for models that do not use a GPU or does not have CUDA version set."
                    ),
                )
            if model_meta.env.cuda_version:
                model_meta_deploy.env.generate_env_for_cuda()
        else:
            # If user does not need GPU, we set this copies cuda_version to None, thus when Image builder gets a
            # not-None cuda_version, it gets to know that GPU is required.
            model_meta_deploy.env._cuda_version = None

        _validate_compute_pool(session, options=options)

        # TODO[shchen]: SNOW-863701, Explore ways to prevent entire model zip being downloaded during deploy step
        #  (for both warehouse and snowservice deployment)
        # One alternative is for model registry to duplicate the model metadata and env dependency storage from model
        # zip so that we don't have to pull down the entire model zip.
        ss_deployment = SnowServiceDeployment(
            session=session,
            model_id=model_id,
            model_meta=model_meta_deploy,
            service_func_name=service_func_name,
            model_zip_stage_path=model_zip_stage_path,  # Pass down model_zip_stage_path for service spec file
            deployment_stage_path=deployment_stage_path,
            target_method=target_method,
            options=options,
        )
        return ss_deployment.deploy()
    finally:
        session.sql(f"ALTER SESSION SET PYTHON_CONNECTOR_QUERY_RESULT_FORMAT = '{prev_format}'").collect()
        # Preserve the original logging level.
        snowpark_logger.setLevel(snowpark_log_level)
        snowflake_connector_logger.setLevel(snowflake_connector_log_level)


def _validate_compute_pool(session: Session, *, options: deploy_options.SnowServiceDeployOptions) -> None:
    # Remove full qualified name to avoid double quotes, which does not work well in desc compute pool syntax.
    compute_pool = options.compute_pool.replace('"', "")
    sql = f"DESC COMPUTE POOL {compute_pool}"
    result = (
        query_result_checker.SqlResultValidator(
            session=session,
            query=sql,
        )
        .has_column("instance_family")
        .has_column("state")
        .has_column("auto_resume")
        .has_dimensions(expected_rows=1)
        .validate()
    )

    state = result[0]["state"]
    auto_resume = bool(result[0]["auto_resume"])

    if state == "SUSPENDED":
        if not auto_resume:
            raise snowml_exceptions.SnowflakeMLException(
                error_code=error_codes.INVALID_SNOWPARK_COMPUTE_POOL,
                original_exception=RuntimeError(
                    "The compute pool you are requesting to use is suspended without auto-resume enabled"
                ),
            )

    elif state not in ["ACTIVE", "IDLE"]:
        raise snowml_exceptions.SnowflakeMLException(
            error_code=error_codes.INVALID_SNOWPARK_COMPUTE_POOL,
            original_exception=RuntimeError(
                "The compute pool you are requesting to use is not in the ACTIVE/IDLE status."
            ),
        )

    if options.use_gpu:
        assert options.num_gpus is not None
        request_gpus = options.num_gpus
        instance_family = result[0]["instance_family"]
        if instance_family in instance_types.INSTANCE_TYPE_TO_GPU_COUNT:
            gpu_capacity = instance_types.INSTANCE_TYPE_TO_GPU_COUNT[instance_family]
            if request_gpus > gpu_capacity:
                raise snowml_exceptions.SnowflakeMLException(
                    error_code=error_codes.INVALID_SNOWPARK_COMPUTE_POOL,
                    original_exception=RuntimeError(
                        f"GPU request exceeds instance capability; {instance_family} instance type has total "
                        f"capacity of {gpu_capacity} GPU, yet a request was made for {request_gpus} GPUs."
                    ),
                )
        else:
            logger.warning(f"Unknown instance type: {instance_family}, skipping GPU validation")


def _get_or_create_image_repo(session: Session, *, service_func_name: str, image_repo: Optional[str] = None) -> str:
    def _sanitize_dns_url(url: str) -> str:
        # Align with existing SnowService image registry url standard.
        return url.lower()

    if image_repo:
        return _sanitize_dns_url(image_repo)

    try:
        conn = session._conn._conn
        # We try to use the same db and schema as the service function locates, as we could retrieve those information
        # if that is a fully qualified one. If not we use the current session one.
        (_db, _schema, _, _) = identifier.parse_schema_level_object_identifier(service_func_name)
        db = _db if _db is not None else conn._database
        schema = _schema if _schema is not None else conn._schema
        assert isinstance(db, str) and isinstance(schema, str)

        client = snowservice_client.SnowServiceClient(session)
        client.create_image_repo(identifier.get_schema_level_object_identifier(db, schema, constants.SNOWML_IMAGE_REPO))
        sql = f"SHOW IMAGE REPOSITORIES LIKE '{constants.SNOWML_IMAGE_REPO}' IN SCHEMA {'.'.join([db, schema])}"
        result = (
            query_result_checker.SqlResultValidator(
                session=session,
                query=sql,
            )
            .has_column("repository_url")
            .has_dimensions(expected_rows=1)
            .validate()
        )
        repository_url = result[0]["repository_url"]
        return str(repository_url)
    except Exception as e:
        raise snowml_exceptions.SnowflakeMLException(
            error_code=error_codes.INTERNAL_SNOWPARK_CONTAINER_SERVICE_ERROR,
            original_exception=RuntimeError("Failed to retrieve image repo URL"),
        ) from e


class SnowServiceDeployment:
    """
    Class implementation that encapsulates image build and workflow deployment to SnowService
    """

    def __init__(
        self,
        session: Session,
        model_id: str,
        model_meta: model_meta.ModelMetadata,
        service_func_name: str,
        model_zip_stage_path: str,
        deployment_stage_path: str,
        target_method: str,
        options: deploy_options.SnowServiceDeployOptions,
    ) -> None:
        """Initialization

        Args:
            session: Snowpark session
            model_id: Unique hex string of length 32, provided by model registry; if not provided, auto-generate one for
                        resource naming.The model_id serves as an idempotent key throughout the deployment workflow.
            model_meta: Model Metadata.
            service_func_name: The service function name in SnowService associated with the created service.
            model_zip_stage_path: Path to model zip file in stage.
            deployment_stage_path: Path to stage containing deployment artifacts.
            target_method: The name of the target method to be deployed.
            options: A SnowServiceDeployOptions object containing deployment options.
        """

        self.session = session
        self.id = model_id
        self.model_meta = model_meta
        self.service_func_name = service_func_name
        self.model_zip_stage_path = model_zip_stage_path
        self.options = options
        self.target_method = target_method
        (db, schema, _, _) = identifier.parse_schema_level_object_identifier(service_func_name)

        self._service_name = identifier.get_schema_level_object_identifier(db, schema, f"service_{model_id}")
        # Spec file and future deployment related artifacts will be stored under {stage}/models/{model_id}
        self._model_artifact_stage_location = posixpath.join(deployment_stage_path, "models", self.id)
        self.debug_dir: Optional[str] = None
        if self.options.debug_mode:
            self.debug_dir = tempfile.mkdtemp()
            logger.warning(f"Debug model is enabled, deployment artifacts will be available in {self.debug_dir}")

    def deploy(self) -> type_hints.SnowparkContainerServiceDeployDetails:
        """
        This function triggers image build followed by workflow deployment to SnowService.

        Returns:
            Deployment details.
        """
        if self.options.prebuilt_snowflake_image:
            logger.warning(f"Skipped image build. Use prebuilt image: {self.options.prebuilt_snowflake_image}")
            service_function_sql = self._deploy_workflow(self.options.prebuilt_snowflake_image)
        else:
            with _debug_aware_tmp_directory(debug_dir=self.debug_dir) as context_dir:
                extra_kwargs = {}
                if self.options.model_in_image:
                    extra_kwargs = {
                        "session": self.session,
                        "model_zip_stage_path": self.model_zip_stage_path,
                    }
                dc = docker_context.DockerContext(
                    context_dir=context_dir,
                    model_meta=self.model_meta,
                    **extra_kwargs,  # type: ignore[arg-type]
                )
                dc.build()
                image_repo = _get_or_create_image_repo(
                    self.session, service_func_name=self.service_func_name, image_repo=self.options.image_repo
                )
                full_image_name = self._get_full_image_name(image_repo=image_repo, context_dir=context_dir)
                registry_client = image_registry_client.ImageRegistryClient(self.session, full_image_name)

                if not self.options.force_image_build and registry_client.image_exists(full_image_name=full_image_name):
                    logger.warning(
                        f"Similar environment detected. Using existing image {full_image_name} to skip image "
                        f"build. To disable this feature, set 'force_image_build=True' in deployment options"
                    )
                else:
                    logger.warning(
                        "Building the Docker image and deploying to Snowpark Container Service. "
                        "This process may take anywhere from a few minutes to a longer period for GPU-based models."
                    )
                    start = time.time()
                    self._build_and_upload_image(
                        context_dir=context_dir, image_repo=image_repo, full_image_name=full_image_name
                    )
                    end = time.time()
                    logger.info(f"Time taken to build and upload image to registry: {end - start:.2f} seconds")
                    logger.warning(
                        f"Image successfully built! For future model deployments, the image will be reused if "
                        f"possible, saving model deployment time. To enforce using the same image, include "
                        f"'prebuilt_snowflake_image': '{full_image_name}' in the deploy() function's options."
                    )

                # Adding the model name as an additional tag to the existing image, excluding the version to prevent
                # excessive tags and also due to version not available in current model metadata. This will allow
                # users to associate images with specific models and perform relevant image registry actions. In the
                # event that model dependencies change across versions, a new image hash will be computed, resulting in
                # a new image.
                try:
                    registry_client.add_tag_to_remote_image(
                        original_full_image_name=full_image_name, new_tag=self.model_meta.name
                    )
                except Exception as e:
                    # Proceed to the deployment with a warning message.
                    logger.warning(f"Failed to add tag {self.model_meta.name} to image {full_image_name}: {str(e)}")
                service_function_sql = self._deploy_workflow(full_image_name)

        rows = self.session.sql(f"DESCRIBE SERVICE {self._service_name}").collect()
        service_info = rows[0].as_dict() if rows and rows[0] else None
        return type_hints.SnowparkContainerServiceDeployDetails(
            service_info=service_info,
            service_function_sql=service_function_sql,
        )

    def _get_full_image_name(self, image_repo: str, context_dir: str) -> str:
        """Return a valid full image name that consists of image name and tag. e.g
        org-account.registry.snowflakecomputing.com/db/schema/repo/image:latest

        Args:
            image_repo: image repo path, e.g. org-account.registry.snowflakecomputing.com/db/schema/repo
            context_dir: the local docker context directory, which consists everything needed to build the docker image.

        Returns:
            Full image name.
        """
        image_repo = _get_or_create_image_repo(
            self.session, service_func_name=self.service_func_name, image_repo=self.options.image_repo
        )

        # We skip "MODEL_METADATA_FILE" as it contains information that will always lead to cache misses.  This isn't an
        # issue because model dependency is also captured in the model env/ folder, which will be hashed. The aim is to
        # reuse the same Docker image even if the user logs a similar model without new dependencies.
        docker_context_dir_hash = file_utils.hash_directory(
            context_dir, ignore_hidden=True, excluded_files=[model_meta.MODEL_METADATA_FILE]
        )
        # By default, we associate a 'latest' tag with each of our created images for easy existence checking.
        # Additional tags are added for readability.
        return f"{image_repo}/{docker_context_dir_hash}:{constants.LATEST_IMAGE_TAG}"

    def _build_and_upload_image(self, context_dir: str, image_repo: str, full_image_name: str) -> None:
        """Handles image build and upload to image registry.

        Args:
            context_dir: the local docker context directory, which consists everything needed to build the docker image.
            image_repo: image repo path, e.g. org-account.registry.snowflakecomputing.com/db/schema/repo
            full_image_name: Full image name consists of image name and image tag.
        """
        image_builder: base_image_builder.ImageBuilder
        if self.options.enable_remote_image_build:
            image_builder = server_image_builder.ServerImageBuilder(
                context_dir=context_dir,
                full_image_name=full_image_name,
                image_repo=image_repo,
                session=self.session,
                artifact_stage_location=self._model_artifact_stage_location,
                compute_pool=self.options.compute_pool,
            )
        else:
            image_builder = client_image_builder.ClientImageBuilder(
                context_dir=context_dir, full_image_name=full_image_name, image_repo=image_repo, session=self.session
            )
        image_builder.build_and_upload_image()

    def _prepare_and_upload_artifacts_to_stage(self, image: str) -> None:
        """Constructs and upload service spec to stage.

        Args:
            image: Name of the image to create SnowService container from.
        """
        if self.options.model_in_image:
            spec_template = (
                importlib_resources.files(snowservice)
                .joinpath("templates/service_spec_template_with_model")  # type: ignore[no-untyped-call]
                .read_text("utf-8")
            )
        else:
            spec_template = (
                importlib_resources.files(snowservice)
                .joinpath("templates/service_spec_template")  # type: ignore[no-untyped-call]
                .read_text("utf-8")
            )

        with _debug_aware_tmp_directory(self.debug_dir) as dir_path:
            spec_file_path = os.path.join(dir_path, f"{constants.SERVICE_SPEC}.yaml")

            with open(spec_file_path, "w+", encoding="utf-8") as spec_file:
                assert self.model_zip_stage_path.startswith("@")
                norm_stage_path = posixpath.normpath(identifier.remove_prefix(self.model_zip_stage_path, "@"))
                # Ensure model stage path has root prefix as stage mount will it mount it to root.
                absolute_model_stage_path = os.path.join("/", norm_stage_path)
                (db, schema, stage, path) = identifier.parse_schema_level_object_identifier(norm_stage_path)
                substitutes = {
                    "image": image,
                    "predict_endpoint_name": constants.PREDICT,
                    "model_stage": identifier.get_schema_level_object_identifier(db, schema, stage),
                    "model_zip_stage_path": absolute_model_stage_path,
                    "inference_server_container_name": constants.INFERENCE_SERVER_CONTAINER,
                    "target_method": self.target_method,
                    "num_workers": self.options.num_workers,
                    "use_gpu": self.options.use_gpu,
                    "enable_ingress": self.options.enable_ingress,
                }
                if self.options.model_in_image:
                    del substitutes["model_stage"]
                    del substitutes["model_zip_stage_path"]
                content = string.Template(spec_template).substitute(substitutes)
                content_dict = yaml.safe_load(content)
                if self.options.use_gpu:
                    container = content_dict["spec"]["container"][0]
                    # TODO[shchen]: SNOW-871538, external dependency that only single GPU is supported on SnowService.
                    # GPU limit has to be specified in order to trigger the workload to be run on GPU in SnowService.
                    container["resources"] = {
                        "limits": {"nvidia.com/gpu": self.options.num_gpus},
                        "requests": {"nvidia.com/gpu": self.options.num_gpus},
                    }

                    # Make LLM use case sequential
                    if any(
                        model_blob_meta.model_type == "huggingface_pipeline" or model_blob_meta.model_type == "llm"
                        for model_blob_meta in self.model_meta.models.values()
                    ):
                        container["env"]["_CONCURRENT_REQUESTS_MAX"] = 1

                yaml.dump(content_dict, spec_file)
                logger.debug("Create service spec: \n, %s", content_dict)

            self.session.file.put(
                local_file_name=spec_file_path,
                stage_location=self._model_artifact_stage_location,
                auto_compress=False,
                overwrite=True,
            )
            logger.debug(
                f"Uploaded spec file {os.path.basename(spec_file_path)} " f"to {self._model_artifact_stage_location}"
            )

    def _get_max_batch_rows(self) -> Optional[int]:
        # To avoid too large batch in HF LLM case
        max_batch_rows = None
        if self.options.use_gpu:
            for model_blob_meta in self.model_meta.models.values():
                batch_size = None
                if model_blob_meta.model_type == "huggingface_pipeline":
                    model_blob_options_hf = cast(
                        model_meta_schema.HuggingFacePipelineModelBlobOptions, model_blob_meta.options
                    )
                    batch_size = model_blob_options_hf["batch_size"]
                if model_blob_meta.model_type == "llm":
                    model_blob_options_llm = cast(model_meta_schema.LLMModelBlobOptions, model_blob_meta.options)
                    batch_size = model_blob_options_llm["batch_size"]
                if batch_size:
                    if max_batch_rows is None:
                        max_batch_rows = batch_size
                    else:
                        max_batch_rows = min(batch_size, max_batch_rows)
        return max_batch_rows

    def _deploy_workflow(self, image: str) -> str:
        """This function handles workflow deployment to SnowService with the given image.

        Args:
            image: Name of the image to create SnowService container from.

        Returns:
            service function sql
        """

        self._prepare_and_upload_artifacts_to_stage(image)
        client = snowservice_client.SnowServiceClient(self.session)
        spec_stage_location = posixpath.join(
            self._model_artifact_stage_location.rstrip("/"), f"{constants.SERVICE_SPEC}.yaml"
        )
        client.create_or_replace_service(
            service_name=self._service_name,
            compute_pool=self.options.compute_pool,
            spec_stage_location=spec_stage_location,
            min_instances=self.options.min_instances,
            max_instances=self.options.max_instances,
        )
        logger.info(f"Wait for service {self._service_name} to become ready...")
        client.block_until_resource_is_ready(
            resource_name=self._service_name, resource_type=constants.ResourceType.SERVICE
        )
        logger.info(f"Service {self._service_name} is ready. Creating service function...")

        service_function_sql = client.create_or_replace_service_function(
            service_func_name=self.service_func_name,
            service_name=self._service_name,
            endpoint_name=constants.PREDICT,
            max_batch_rows=self._get_max_batch_rows(),
        )
        logger.info(f"Service function {self.service_func_name} is created. Deployment completed successfully!")
        return service_function_sql



import logging
import warnings

from snowflake import snowpark
from snowflake.snowpark import functions, types


def cast_snowpark_dataframe(df: snowpark.DataFrame) -> snowpark.DataFrame:
    """Cast columns in the dataframe to types that are compatible with tensor.

    It assists FileSet.make() in performing implicit data casting.

    Args:
        df: A snowpark dataframe.

    Returns:
        A snowpark dataframe whose data type has been casted.
    """

    fields = df.schema.fields
    selected_cols = []
    for field in fields:
        src = field.column_identifier.quoted_name
        if isinstance(field.datatype, types.DecimalType):
            if field.datatype.scale:
                dest: types.DataType = types.FloatType()
            else:
                dest = types.LongType()
            selected_cols.append(functions.cast(functions.col(src), dest).alias(src))
        elif isinstance(field.datatype, types.DoubleType):
            dest = types.FloatType()
            selected_cols.append(functions.cast(functions.col(src), dest).alias(src))
        elif isinstance(field.datatype, types.ByteType):
            # Snowpark maps ByteType to BYTEINT, which will not do the casting job when unloading to parquet files.
            # We will use SMALLINT instead until this issue got fixed.
            # Investigate JIRA filed: SNOW-725041
            dest = types.ShortType()
            selected_cols.append(functions.cast(functions.col(src), dest).alias(src))
        elif field.datatype in (types.ShortType(), types.IntegerType(), types.LongType()):
            dest = field.datatype
            selected_cols.append(functions.cast(functions.col(src), dest).alias(src))
        else:
            if field.datatype in (types.DateType(), types.TimestampType(), types.TimeType()):
                logging.warning(
                    "A Column with DATE or TIMESTAMP data type detected. "
                    "It might not be able to get converted to tensors. "
                    "Please consider handle it in feature engineering."
                )
            elif (
                isinstance(field.datatype, types.ArrayType)
                or isinstance(field.datatype, types.MapType)
                or isinstance(field.datatype, types.VariantType)
            ):
                logging.warning(
                    "A Column with semi-structured data type (variant, array or object) was detected. "
                    "It might not be able to get converted to tensors. "
                    "Please consider handling it in feature engineering."
                )
            selected_cols.append(functions.col(src))
    df = df.select(selected_cols)
    return df


def cast_snowpark_dataframe_column_types(df: snowpark.DataFrame) -> snowpark.DataFrame:
    """Cast columns in the dataframe to types that are compatible with pandas DataFrame.

    It assists modeling API (fit, predict, ...) in performing implicit data casting.
    The reason for casting: snowpark dataframe would transform as pandas dataframe
        to compute within sproc.

    Args:
        df: A snowpark dataframe.

    Returns:
        A snowpark dataframe whose data type has been casted.
    """
    fields = df.schema.fields
    selected_cols = []
    for field in fields:
        src = field.column_identifier.quoted_name
        # Handle DecimalType: Numbers up to 38 digits, with an optional precision and scale
        # By default, precision is 38 and scale is 0 (i.e. NUMBER(38, 0))
        if isinstance(field.datatype, types.DecimalType):
            # If datatype has scale; convert into float/double type
            # In snowflake, DOUBLE is the same as FLOAT, provides precision up to 18.
            if field.datatype.scale:
                dest_dtype: types.DataType = types.DoubleType()
                warnings.warn(
                    f"Warning: The Decimal({field.datatype.precision}, {field.datatype.scale}) data type"
                    " is being automatically converted to DoubleType in the Snowpark DataFrame. "
                    "This automatic conversion may lead to potential precision loss and rounding errors. "
                    "If you wish to prevent this conversion, you should manually perform "
                    "the necessary data type conversion."
                )
            else:
                # IntegerType default as NUMBER(38, 0), but
                # snowpark dataframe would automatically transform to LongType in function `convert_sf_to_sp_type`
                # To align with snowpark, set all the decimal without scale as LongType
                dest_dtype = types.LongType()
                warnings.warn(
                    f"Warning: The Decimal({field.datatype.precision}, 0) data type"
                    " is being automatically converted to LongType in the Snowpark DataFrame. "
                    "This automatic conversion may lead to potential precision loss and rounding errors. "
                    "If you wish to prevent this conversion, you should manually perform "
                    "the necessary data type conversion."
                )
            selected_cols.append(functions.cast(functions.col(src), dest_dtype).alias(src))
        # TODO: add more type handling or error message
        else:
            selected_cols.append(functions.col(src))
    df = df.select(selected_cols)
    return df
